import sys

import numpy as np
import torch as torch
from gpytorch.kernels import Kernel
from linear_operator.operators import KernelLinearOperator
from pykeops.torch import LazyTensor
from scipy.optimize import minimize
from scipy.special import comb, logsumexp
from torch.distributions.transforms import CorrCholeskyTransform
from torch.distributions import Wishart, Beta
from torch.nn import Parameter
from torch.linalg import cholesky

from epik.utils import (
    HammingDistanceCalculator,
    KrawtchoukPolynomials,
    inner_product,
    log1mexp,
    cov2corr
)


class SequenceKernel(Kernel):
    """
    A kernel class for sequence data, inheriting from the base `Kernel` class.

    This class implements methods for calculating kernel matrices and 
    Hamming distances for sequence data, with optional support for 
    KeOps for efficient computation on large datasets.
    
    Parameters
    ----------
    n_alleles : int
        The number of alleles in the sequence data.
    seq_length : int
        The length of the sequences.
    use_keops : bool, optional
        Whether to use KeOps for kernel computation (default is False).
    **kwargs : dict
        Additional keyword arguments passed to the base `Kernel` class.
    """
    def __init__(self, n_alleles, seq_length, use_keops=False, **kwargs):
        self.n_alleles = n_alleles
        self.seq_length = seq_length
        self.lp1 = seq_length + 1
        self.n_features = seq_length * n_alleles

        self.logn = self.seq_length * np.log(self.n_alleles)
        self.logam1 = np.log(self.n_alleles - 1)
        self.use_keops = use_keops
        super().__init__(**kwargs)

    def select_site(self, x, site):
        idx = torch.arange(site * self.n_alleles, (site + 1) * self.n_alleles)
        return x.index_select(-1, idx.to(device=x.device))

    def register_params(self, params={}, constraints={}):
        self.params = params
        for param_name, param in params.items():
            self.register_parameter(name=param_name, parameter=param)

        for param_name, constraint in constraints.items():
            self.register_constraint(param_name, constraint)

    def calc_hamming_distance(self, x1, x2, diag=False, keops=False):
        if diag or not keops:
            s = inner_product(x1, x2, diag=diag)
            d = float(self.seq_length) - s
        else:
            x1_ = LazyTensor(x1[..., :, None, :])
            x2_ = LazyTensor(x2[..., None, :, :])
            s = (x1_ * x2_).sum(-1)
            d = float(self.seq_length) - s
        return d

    def forward(self, x1, x2, diag=False, **kwargs):
        if diag:
            kernel = self._nonkeops_forward(x1, x2, diag=True, **kwargs)

        else:
            if self.use_keops:
                kernel = self._keops_forward(x1, x2, **kwargs)

            else:
                try:
                    kernel = self._nonkeops_forward(x1, x2, diag=False, **kwargs)

                except RuntimeError as error:  # Memory error
                    msg = "\n{}. Likely due to memory error when loading ".format(error)
                    msg += "kernel matrix: switching to KeOps\n"
                    sys.stderr.write(msg)

                    allocated_memory = torch.cuda.memory_allocated(device="cuda")
                    reserved_memory = torch.cuda.memory_reserved(device="cuda")
                    n1, n2 = x1.shape[0], x2.shape[0]
                    sys.stderr.write(
                        f"Kernel matrix memory {(n1, n2)}: {n1 * n2 * x2.element_size() / 1e6} MB"
                    )
                    sys.stderr.write(
                        f"Memory allocated: {allocated_memory / 1e6:.2f} MB\n"
                    )
                    sys.stderr.write(
                        f"Memory reserved: {reserved_memory / 1e6:.2f} MB\n"
                    )
                    sys.stderr.write(torch.cuda.memory_summary(device="cuda"))
                    self.use_keops = True
                    torch.cuda.empty_cache()
                    kernel = self._keops_forward(x1, x2, **kwargs)

        torch.cuda.empty_cache()
        return kernel


class BaseVarianceComponentKernel(SequenceKernel):
    is_stationary = True

    def __init__(self, n_alleles, seq_length, log_lambdas0=None, **kwargs):
        super().__init__(n_alleles, seq_length, **kwargs)
        self.log_lambdas0 = log_lambdas0
        self.ws = KrawtchoukPolynomials(n_alleles, seq_length, max_k=self.max_k)
        self.set_params()

    def get_log_lambdas0(self):
        if self.log_lambdas0 is None:
            log_lambdas0 = torch.zeros(self.max_k + 1)
            # else:
            #     b = self.cov0[1] / (self.cov0[0] - self.cov0[1])
            #     beta0 = np.log(self.cov0[0]) + self.seq_length * (
            #         np.log(1 + self.n_alleles * b) - np.log(1 + b)
            #     )
            #     beta1 = np.log(1 + self.n_alleles * b)
            #     k = torch.arange(self.ws.max_k + 1).to(dtype=torch.float)
            #     log_lambdas0 = beta0 - beta1 * k
        else:
            log_lambdas0 = self.log_lambdas0

        log_lambdas0 = log_lambdas0.to(dtype=self.dtype)
        return log_lambdas0

    def set_params(self):
        c_bk = self.calc_c_bk()
        log_lambdas0 = self.get_log_lambdas0()
        theta = torch.Tensor([[-np.log(0.1) / self.seq_length]])
        params = {
            "log_lambdas": Parameter(log_lambdas0, requires_grad=True),
            "theta": Parameter(theta, requires_grad=False),
            "c_bk": Parameter(c_bk, requires_grad=False),
        }
        self.register_params(params)

    def calc_c_b(self, log_lambdas):
        c_b = self.c_bk @ torch.exp(log_lambdas)
        return c_b

    def get_c_b(self):
        return self.calc_c_b(self.log_lambdas)

    def get_basis(self, d, keops=False):
        if keops:
            yield (1.0)
        else:
            yield (torch.ones_like(d))

        d_power = d
        for _ in range(1, self.max_k):
            yield (d_power)
            d_power = d_power * d
        yield (d_power)

    def d_to_cov(self, d, c_b, keops=False):
        basis = self.get_basis(d, keops=keops)
        b0 = next(basis)
        kernel = c_b[0] * b0
        for b_i, c_i in zip(basis, c_b[1:]):
            kernel += c_i * b_i
        return kernel

    def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
        c_b = self.get_c_b()
        d = self.calc_hamming_distance(x1, x2, diag=diag)
        return self.d_to_cov(d, c_b, keops=False)

    def _covar_func(self, x1, x2, c_b, **kwargs):
        d = self.calc_hamming_distance(x1, x2, keops=True)
        return self.d_to_cov(d, c_b, keops=True)

    def _keops_forward(self, x1, x2, **kwargs):
        c_b = self.get_c_b()
        kernel = KernelLinearOperator(
            x1, x2, covar_func=self._covar_func, c_b=c_b, **kwargs
        )
        return kernel


class AdditiveKernel(BaseVarianceComponentKernel):
    r"""
    Additive kernel for functions on sequence space.

    This kernel computes the covariance between two sequences as a linear function
    of the Hamming distance separating them. The parameters are derived from the
    variance contributions of the constant and additive components.

    .. math::
        K(x, y) = c_0 + c_1 \cdot d(x, y)

    where:
    
    .. math::
        c_0 = \lambda_0 + \ell \cdot (\alpha - 1) \cdot \lambda_1
        
    .. math::
        c_1 = -\alpha \cdot \lambda_1

    Here, :math:`\lambda_0` and :math:`\lambda_1` are variance parameters, :math:`\ell` is the sequence
    length, and :math:`\alpha` is the number of alleles.

    When applied to one-hot encoded sequence embeddings :math:`x_1` and :math:`x_2`, this kernel
    returns a linear operator that facilitates efficient matrix-vector products
    without explicitly constructing the full covariance matrix.
    """

    @property
    def max_k(self):
        return 1

    def calc_c_bk(self):
        a, sl = self.n_alleles, self.seq_length
        c_bk = torch.tensor([[1.0, sl * (a - 1)], [0, -a]])
        return c_bk

    def forward(self, x1, x2, diag=False, **kwargs):
        c_b = self.get_c_b()
        if diag:
            d = self.calc_hamming_distance(x1, x2, diag=True)
            kernel = c_b[0] + c_b[1] * d
        else:
            calc_d = HammingDistanceCalculator(
                self.seq_length, scale=c_b[1], shift=c_b[0]
            )
            kernel = calc_d(x1, x2)
        return kernel


class PairwiseKernel(BaseVarianceComponentKernel):
    """
    Pairwise kernel for functions on sequence space.
     
    The covariance between two sequences is quadratic in the Hamming distance
    that separates them, with coefficients determined by the variance
    explained by the constant, additive and pairwise components.

    .. math::
        K(x, y) = c_0 + c_1 \cdot d(x, y) + c_2 \cdot d(x, y)^2

    These coefficients result from expanding the Krawtchouk polynomials
    of order 2, as in the additive kernel, and allows computing the covariance
    matrix easily for any number of sequences of any length.
    """

    @property
    def max_k(self):
        return 2

    def get_basis(self, d, keops=False):
        b0 = 1.0 if keops else torch.ones_like(d)
        yield (b0)
        yield (d)
        yield (d.square())

    def calc_c_bk(self):
        a, sl = self.n_alleles, self.seq_length
        c13 = (
            a * sl
            - 0.5 * a**2 * sl
            - 0.5 * sl
            - a * sl**2
            + 0.5 * a**2 * sl**2
            + 0.5 * sl**2
        )
        c23 = -a + 0.5 * a**2 + a * sl - a**2 * sl
        c_bk = torch.tensor([[1, sl * (a - 1), c13], [0, -a, c23], [0, 0, 0.5 * a**2]])
        return c_bk
        # c_sb = torch.tensor([[1, sl,         sl ** 2 ],
        #                     [0,  -1,        -2  * sl ],
        #                     [0,   0,               1.]])
        # return(c_sb @ c_bk)

    # def forward(self, x1, x2, diag=False, **kwargs):
    #     c_b = self.get_c_b()
    #     if diag:
    #         # d = self.calc_hamming_distance(x1, x2, diag=True)
    #         # kernel = c_b[0] + c_b[1] * d + c_b[2] * d ** 2
    #         s = inner_product(x1, x2, diag=True)
    #         kernel = c_b[0] + c_b[1] * s + c_b[2] * torch.square(s)
    #     else:
    #         # calc_d = HammingDistanceCalculator(self.seq_length, scale=c_b[1], shift=c_b[0])
    #         # kernel = calc_d(x1, x2)
    #         ones1 = torch.ones((x1.shape[0], 1), device=x1.device)
    #         ones2 = torch.ones((1, x2.shape[0]), device=x1.device)
    #         s0 = MatmulLinearOperator(c_b[0] * ones1, ones2)
    #         s1 = MatmulLinearOperator(c_b[1] * x1, x2.T)
    #         s2 = SquaredMatMulOperator(c_b[2] * x1, x2.T)
    #         kernel = s0 + s1 + s2
    #     return(kernel)


class VarianceComponentKernel(BaseVarianceComponentKernel):
    """
    Variance Component Kernel for functions on sequence space.

    This kernel computes the covariance between two sequences using
    Krawtchouk polynomials.

    .. math::
        K(x, y) = \sum_{k=0}^{\ell} \lambda_k \cdot K_k(x, y)

    To ensure differentiability in PyTorch, the covariance for each
    distance class is precomputed, and a kernel interpolation approach
    is used to compute the covariance between input sequence pairs.
    """
    is_stationary = True

    def __init__(
        self,
        n_alleles,
        seq_length,
        log_lambdas0=None,
        max_k=None,
        **kwargs,
    ):
        self.max_k = max_k if max_k is not None else seq_length
        super().__init__(
            n_alleles,
            seq_length,
            log_lambdas0=log_lambdas0,
            **kwargs,
        )

    def get_basis(self, d, keops=False):
        for i in range(self.seq_length + 1):
            d_i = float(i)
            b_i = (-self.theta[0, 0] * (d - d_i).abs()).exp()
            yield (b_i)

    def calc_c_bk(self):
        return self.ws.c_bk


class SiteProductKernel(SequenceKernel):
    is_stationary = True

    def __init__(
        self,
        n_alleles,
        seq_length,
        log_var0=None,
        theta0=None,
        **kwargs,
    ):
        super().__init__(n_alleles, seq_length, **kwargs)
        self.theta0 = theta0
        self.log_var0 = log_var0
        self.set_params()
        self.site_shape = (self.n_alleles, self.n_alleles)

    def is_positive(self):
        return False

    def calc_log_var0(self):
        if self.log_var0 is None:
            log_var0 = torch.zeros(1)
        else:
            log_var0 = self.log_var0
        return log_var0

    def set_params(self):
        theta = Parameter(self.calc_theta0(), requires_grad=True)
        log_var0 = Parameter(self.calc_log_var0(), requires_grad=True)
        self.register_parameter(name="theta", parameter=theta)
        self.register_parameter(name="log_var", parameter=log_var0)

    def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
        if self.is_positive():
            site_log_kernels = self.get_site_log_kernels()

            if diag:
                min_size = min(x1.shape[0], x2.shape[0])
                log_kernel = 0.0
                for i in range(self.seq_length):
                    log_kernel += (
                        (self.select_site(x1[:min_size], site=i) @ site_log_kernels[i])
                        * self.select_site(x2[:min_size], site=i)
                    ).sum(1)

            else:
                log_kernel = 0
                for i in range(self.seq_length):
                    log_kernel += (
                        self.select_site(x1, site=i)
                        @ site_log_kernels[i]
                        @ self.select_site(x2, site=i).T
                    )

            return torch.exp(self.log_var + log_kernel)

        else:
            site_kernels = self.get_site_kernels()
            sigma2 = torch.exp(self.log_var)

            if diag:
                min_size = min(x1.shape[0], x2.shape[0])
                kernel = 1.0
                for i in range(self.seq_length):
                    kernel *= (
                        (self.select_site(x1, site=i) @ site_kernels[i])
                        * self.select_site(x2, site=i)
                    ).sum(1)

            else:
                kernel = 1.0
                for i in range(self.seq_length):
                    kernel *= (
                        self.select_site(x1, site=i)
                        @ site_kernels[i]
                        @ self.select_site(x2, site=i).T
                    )

            return sigma2 * kernel

    def _covar_func(self, x1, x2, **kwargs):
        x1_ = LazyTensor(self.select_site(x1, site=0)[:, None, :])
        x2_ = LazyTensor(self.select_site(x2, site=0)[None, :, :])
        K = (x1_ * x2_).sum(-1)

        for i in range(1, self.seq_length):
            x1_ = LazyTensor(self.select_site(x1, site=i)[:, None, :])
            x2_ = LazyTensor(self.select_site(x2, site=i)[None, :, :])
            K *= (x1_ * x2_).sum(-1)

        return K

    def _covar_func_log(self, x1, x2, **kwargs):
        x1_ = LazyTensor(x1[:, None, :])
        x2_ = LazyTensor(x2[None, :, :])
        K = (x1_ * x2_).sum(-1).exp()
        return K

    def _keops_forward(self, x1, x2, **kwargs):
        if self.is_positive():
            site_log_kernels = self.get_site_log_kernels()
            M = torch.block_diag(*site_log_kernels)
            sigma2 = torch.exp(self.log_var)
            kernel = sigma2 * KernelLinearOperator(
                x1 @ M, x2, covar_func=self._covar_func_log, **kwargs
            )

        else:
            site_kernels = [x for x in self.get_site_kernels()]
            sigma2 = torch.exp(self.log_var)
            M = torch.block_diag(*site_kernels)
            kernel = sigma2 * KernelLinearOperator(
                x1 @ M, x2, covar_func=self._covar_func, **kwargs
            )
        return kernel

    def get_delta(self):
        """
        Compute the decay factors of the kernel.

        The decay factors represent the percentage decrease in predictability 
        when introducing a specific mutation.

        Returns
        -------
        delta : torch.Tensor
            A tensor containing the decay factors.
        """
        delta = self.theta_to_delta(self.theta, n_alleles=self.n_alleles)
        return delta

    def get_mutation_delta(self):
        """
        Compute the mutation-specific decay factors of the kernel.

        The decay factors represent the percentage decrease in predictability 
        when introducing a specific mutation.

        Returns
        -------
        delta : torch.Tensor
            A tensor containing the decay factors for each possible mutation.
        """
        Ks = self.get_site_kernels()
        return 1 - Ks


class ExponentialKernel(SiteProductKernel):
    r"""
    Exponential Kernel for functions on sequence space.

    This kernel computes the covariance between two sequences as a
    geometrically decaying function of the Hamming distance separating them.


    .. math::
        K(x, y) = \left( \frac{ 1-\rho }{ 1 + (\alpha - 1)\rho } \right)^d

    where:
    
        - :math:`\rho` is a parameter controlling the decay rate.
        
        - :math:`\alpha` is the number of alleles.
        
        - :math:`d` is the Hamming distance between sequences :math:`x` and :math:`y`.

    """
    def get_site_kernel(self):
        rho = torch.exp(self.theta)
        v = (1 - rho) / (1 + (self.n_alleles - 1) * rho)
        kernel = (
            torch.ones((self.n_alleles, self.n_alleles), device=self.theta.device) * v
        ).fill_diagonal_(1.0)
        return kernel

    def get_site_log_kernel(self):
        w = log1mexp(self.theta) - torch.log1p(
            (self.n_alleles - 1) * torch.exp(self.theta)
        )
        log_kernel = (
            torch.ones((self.n_alleles, self.n_alleles), device=self.theta.device) * w
        ).fill_diagonal_(0.0)
        return log_kernel

    def calc_theta0(self):
        if self.theta0 is None:
            q = torch.Tensor([np.exp(-np.log(10) / self.seq_length)])
            qs = Beta(20 * q, 20 * (1 - q)).sample((1, ))
            rho = (1 - qs) / (1 + (self.n_alleles - 1) * qs)
            theta0 = torch.log(rho)
        else:
            theta0 = self.theta0
        return theta0

    def get_site_kernels(self):
        kernel = self.get_site_kernel()
        kernels = torch.stack([kernel] * self.seq_length, axis=0)
        return kernels

    def get_site_log_kernels(self):
        log_kernel = self.get_site_log_kernel()
        log_kernels = torch.stack([log_kernel] * self.seq_length, axis=0)
        return log_kernels

    def theta_to_delta(self, theta, n_alleles):
        rho = torch.exp(theta)
        delta = 1 - (1 - rho) / (1 + (n_alleles - 1) * rho)
        return delta

    def is_positive(self):
        return torch.all(self.theta < 0.0)

    def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
        if self.is_positive():
            w = log1mexp(self.theta) - torch.log1p(
                (self.n_alleles - 1) * torch.exp(self.theta)
            )
            d = self.calc_hamming_distance(x1, x2, diag=diag, keops=False)
            return torch.exp(self.log_var + w * d)

        else:
            return super()._nonkeops_forward(x1, x2, diag=diag, **kwargs)


class ConnectednessKernel(SiteProductKernel):
    r"""
    Connectedness Kernel for functions on sequence space.

    This kernel computes the covariance between two sequences where
    mutations at different sites have different effects on the 
    predictability of other mutations
    

    .. math::
        K(x, y) = \prod_p^{\ell}\frac{1-\rho_p}{1 + (\alpha - 1)\rho_p}

    where:
    
        - :math:`\rho_p` is a parameter controlling the decay rate of site :math:`p`.
        
        - :math:`\alpha` is the number of alleles.
        
        - :math:`\ell` is the sequence length.

    """
    def calc_theta0(self):
        if self.theta0 is None:
            q = torch.Tensor([np.exp(-np.log(10) / self.seq_length)])
            qs = Beta(20 * q, 20 * (1 - q)).sample((self.seq_length, ))
            rho = (1 - qs) / (1 + (self.n_alleles - 1) * qs)
            theta0 = torch.log(rho)
        else:
            theta0 = self.theta0
        return theta0

    def get_site_kernels(self):
        rho = torch.exp(self.theta).reshape((self.seq_length, 1, 1))
        v = (1 - rho) / (1 + (self.n_alleles - 1) * rho)
        kernels = (
            torch.ones(
                (self.seq_length, self.n_alleles, self.n_alleles),
                device=self.theta.device,
            )
            * v
        )
        for i in range(self.seq_length):
            kernels[i].fill_diagonal_(1.0)
        return kernels

    def get_site_log_kernels(self):
        ws = log1mexp(self.theta) - torch.log1p(
            (self.n_alleles - 1) * torch.exp(self.theta)
        )
        log_kernels = []
        size = (self.n_alleles, self.n_alleles)
        for w in ws:
            log_kernel = w * torch.ones(size, device=self.theta.device)
            log_kernels.append(log_kernel.fill_diagonal_(0.0))
        return torch.stack(log_kernels, axis=0)

    def is_positive(self):
        return torch.all(self.theta < 0.0)

    def theta_to_delta(self, theta, n_alleles):
        rho = torch.exp(theta)
        delta = 1 - (1 - rho) / (1 + (n_alleles - 1) * rho)
        return delta


class JengaKernel(SiteProductKernel):
    r"""
    Jenga Kernel for functions on sequence space.

    This kernel computes the covariance between two sequences as the product 
    of allele- and site-specific factors at the alleles where they differ.

    .. math::
        K(x, y) = \prod_{p: x_p \neq y_p} 
        \sqrt{\frac{1-\rho_p}{1 + \frac{1-\pi_p^{x_p}}{\pi_p^{x_p}}\rho_p}}
        \sqrt{\frac{1-\rho_p}{1 + \frac{1-\pi_p^{y_p}}{\pi_p^{y_p}}\rho_p}}

    where:
    
        - :math:`\rho_p` is a parameter controlling the decay rate at site :math:`p`.
        
        - :math:`\pi_p^{x_p}` and :math:`\pi_p^{y_p}` are site and allele specific probabilities.
        
        - :math:`\ell` is the sequence length.
    """

    def calc_theta0(self):
        if self.theta0 is None:
            q = torch.Tensor([np.exp(-np.log(10) / self.seq_length)])
            qs = Beta(20 * q, 20 * (1 - q)).sample((self.seq_length, ))
            rho = (1 - qs) / (1 + (self.n_alleles - 1) * qs)
            theta0 = torch.randn((self.seq_length, self.n_alleles + 1))
            theta0[:, :1] = torch.log(rho)
        else:
            theta0 = self.theta0
        return theta0

    def get_log_rho_log1p_eta_rho(self, theta):
        log_rho = theta[:, 0].unsqueeze(1)
        log_p = theta[:, 1:] - torch.logsumexp(theta[:, 1:], 1).unsqueeze(1)
        log_eta = log1mexp(log_p) - log_p

        log1p_eta_rho = torch.logaddexp(torch.zeros_like(log_eta), log_rho + log_eta)
        return log_rho, log1p_eta_rho

    def get_site_kernels(self):
        log_rho, log1p_eta_rho = self.get_log_rho_log1p_eta_rho(self.theta)
        rho = torch.exp(log_rho).reshape((self.seq_length, 1, 1))
        allele_factors = 0.5 * log1p_eta_rho
        denom = torch.exp(-allele_factors.unsqueeze(1) - allele_factors.unsqueeze(2))
        kernel = (1 - rho) * denom
        for i in range(self.seq_length):
            kernel[i].fill_diagonal_(1.0)
        return kernel

    def get_site_log_kernels(self):
        log_rho, log1p_eta_rho = self.get_log_rho_log1p_eta_rho(self.theta)
        log1m_rho = log1mexp(log_rho)
        zs = 0.5 * (log1m_rho - log1p_eta_rho)

        log_kernels = []
        for z in zs:
            log_kernel = z.unsqueeze(0) + z.unsqueeze(1)
            log_kernels.append(log_kernel.fill_diagonal_(0.0))
        return torch.stack(log_kernels, axis=0)

    def is_positive(self):
        return torch.all(self.theta[:, 0] < 0.0)

    def theta_to_delta(self, theta, **kwargs):
        log_rho, log1p_eta_rho = self.get_log_rho_log1p_eta_rho(theta)
        onemrho = 1 - torch.exp(log_rho)
        site_factors = torch.exp(-0.5 * log1p_eta_rho)
        delta = 1 - torch.sign(onemrho) * torch.sqrt(torch.abs(onemrho)) * site_factors
        return delta


class GeneralProductKernel(SiteProductKernel):
    """
    General Product Kernel for sequence data.

    This kernel computes the covariance between two sequences as the product
    of site-specific kernels, where each site kernel is parameterized by the
    Cholesky factor of a correlation matrix.

    .. math::
        K(x, y) = \prod_{p=1}^\ell K_p(x_p, y_p),

    where:
    
    .. math::
        K_p = L L^T,

    and :math:`L` is the Cholesky factor of the correlation matrix, 
    parameterized using the LKJ transform.
    """
    is_stationary = True

    def __init__(self, n_alleles, seq_length, theta0=None, **kwargs):
        self.dim = int(comb(n_alleles, 2))
        self.theta_to_L = CorrCholeskyTransform()
        super().__init__(n_alleles, seq_length, theta0=theta0, **kwargs)

    def calc_theta0(self):
        if self.theta0 is not None:
            theta0 = self.theta0
        else:
            q = torch.Tensor([np.exp(-np.log(10) / self.seq_length)])
            C = (1 - q) * torch.eye(self.n_alleles) + q * torch.ones(
                (self.n_alleles, self.n_alleles)
            )
            df = 21.
            Ls = [cholesky(cov2corr(Wishart(torch.Tensor([df]), C).sample()[0]))
                  for _ in range(self.seq_length)]
            theta0 = torch.stack([self.theta_to_L._inverse(L)
                                  for L in Ls], axis=0)
        return theta0

    def theta_to_cor(self, theta):
        seq_length = theta.shape[0]
        Ls = [self.theta_to_L(theta[i]) for i in range(seq_length)]
        return torch.stack([(L @ L.T) for L in Ls], axis=0)

    def get_site_kernels(self):
        return self.theta_to_cor(self.theta)
    
    def get_site_log_kernels(self):
        return torch.log(self.get_site_kernels())

    def is_positive(self):
        return torch.all(self.get_site_kernels() > 0.0)

    def theta_to_delta(self, theta, **kwargs):
        return 1 - self.theta_to_cor(theta)

    def get_delta(self):
        return self.get_mutation_delta()


def get_kernel(
    kernel, n_alleles, seq_length, theta0=None, log_var0=None, log_lambdas0=None
):
    kernels = {
        "Additive": AdditiveKernel,
        "Pairwise": PairwiseKernel,
        "VC": VarianceComponentKernel,
        "Exponential": ExponentialKernel,
        "Connectedness": ConnectednessKernel,
        "Jenga": JengaKernel,
        "GeneralProduct": GeneralProductKernel,
    }
    kernel = kernels[kernel](
        n_alleles,
        seq_length,
        theta0=theta0,
        log_var0=log_var0,
        log_lambdas0=log_lambdas0,
    )
    return kernel


class SiteKernelAligner(object):
    def __init__(self, n_alleles, seq_length):
        self.n_alleles = n_alleles
        self.seq_length = seq_length
        self.size = (n_alleles, n_alleles)
        self.n_offdiag = n_alleles**2 - n_alleles
        self.theta_to_L = CorrCholeskyTransform()

    def calc_frob(self, A, B):
        return np.square(A - B).sum()

    def log_rho_to_q(self, log_rho):
        rho = np.exp(log_rho)
        v = (1 - rho) / (1 + (self.n_alleles - 1) * rho)
        return v

    def q_to_log_rho(self, q):
        log_rho = np.log((1 - q) / (1 + (self.n_alleles - 1) * q))
        return log_rho

    def rho_to_corr(self, theta):
        v = self.log_rho_to_q(theta)
        corr = np.full(self.size, v)
        np.fill_diagonal(corr, 1)
        return corr

    def corr_to_q(self, corr):
        q = (corr.sum() - np.diag(corr).sum()) / self.n_offdiag
        return q

    def corr_to_general_product(self, corr):
        L = torch.Tensor(np.linalg.cholesky(corr))
        return self.theta_to_L._inverse(L).numpy()

    def general_product_to_corr(self, theta):
        L = self.theta_to_L(torch.Tensor(theta))
        return (L @ L.T).numpy()

    def exponential_to_connectedness(self, theta):
        return np.vstack([theta] * self.seq_length)

    def exponential_to_jenga(self, theta):
        col1 = np.vstack([theta] * self.seq_length)
        return np.hstack([col1, np.zeros((self.seq_length, self.n_alleles))])

    def exponential_to_general_product(self, theta):
        corr = self.rho_to_corr(theta[0])
        theta = np.vstack([self.corr_to_general_product(corr)] * self.seq_length)
        return theta

    def connectedness_to_exponential(self, theta):
        q = np.mean([self.log_rho_to_q(theta_i) for theta_i in theta])
        theta = np.array([[self.q_to_log_rho(q)]])
        return theta

    def connectedness_to_jenga(self, theta):
        z = np.zeros((self.seq_length, self.n_alleles))
        return np.hstack([theta, z])

    def connectedness_to_general_product(self, theta):
        theta = np.vstack(
            [
                self.corr_to_general_product(self.rho_to_corr(theta_i))
                for theta_i in theta
            ]
        )
        return theta

    def jenga_to_corr(self, theta):
        log_rho, log_p = theta[0], theta[1:]
        rho = np.exp(log_rho)
        p = np.exp(log_p - logsumexp(log_p))
        eta = (1 - p) / p
        fs = np.sqrt(1 + eta * rho)
        corr = (1 - rho.reshape((1, 1))) / (
            np.expand_dims(fs, 0) * np.expand_dims(fs, 1)
        )
        np.fill_diagonal(corr, 1)
        return corr

    def jenga_to_connectedness(self, theta):
        qs = [self.corr_to_q(self.jenga_to_corr(theta_i)) for theta_i in theta]
        theta = np.array([[self.q_to_log_rho(q)] for q in qs])
        return theta

    def jenga_to_exponential(self, theta):
        q = np.mean([self.corr_to_q(self.jenga_to_corr(theta_i)) for theta_i in theta])
        theta = np.array([[self.q_to_log_rho(q)]])
        return theta

    def jenga_to_general_product(self, theta):
        theta = np.vstack(
            [
                self.corr_to_general_product(self.jenga_to_corr(theta_i))
                for theta_i in theta
            ]
        )
        return theta

    def general_product_to_exponential(self, theta):
        q = np.mean(
            [self.corr_to_q(self.general_product_to_corr(theta_i)) for theta_i in theta]
        )
        theta = np.array([[self.q_to_log_rho(q)]])
        return theta

    def general_product_to_connectedness(self, theta):
        qs = [
            self.corr_to_q(self.general_product_to_corr(theta_i)) for theta_i in theta
        ]
        theta = np.array([[self.q_to_log_rho(q)] for q in qs])
        return theta

    def general_product_to_jenga(self, theta):
        thetas = []
        for theta_i in theta:
            B = self.general_product_to_corr(theta_i)

            def loss(params):
                A = self.jenga_to_corr(params)
                return self.calc_frob(A, B)

            params0 = np.zeros(self.n_alleles + 1)
            res = minimize(loss, x0=params0)
            thetas.append(res.x)
        return np.vstack(thetas)


class FactorAnalysisKernel(Kernel):
    def __init__(self, n_alleles, seq_length, ndim, **kwargs):
        self.n_alleles = n_alleles
        self.seq_length = seq_length
        self.ndim = ndim
        self.nfeatures = seq_length * n_alleles
        super().__init__(**kwargs)

        q_raw0 = torch.normal(0, 1, size=(self.nfeatures, self.ndim))
        q_raw = Parameter(q_raw0, requires_grad=True)

        log_lambdas0_sqrt0 = torch.linspace(0, -2, self.ndim).unsqueeze(0)
        log_lambdas0_sqrt = Parameter(log_lambdas0_sqrt0, requires_grad=True)
        log_sigma2 = Parameter(torch.Tensor([0.]), requires_grad=True)

        self.register_parameter(name="log_lambdas_sqrt", parameter=log_lambdas0_sqrt)
        self.register_parameter(name="q_raw", parameter=q_raw)
        self.register_parameter(name="log_sigma2", parameter=log_sigma2)

    @property
    def lambdas_sqrt(self):
        return(torch.exp(self.log_lambdas_sqrt))

    @property
    def q(self):
        return(torch.linalg.qr(self.q_raw)[0])
    
    @property
    def sigma2(self):
        return torch.exp(self.log_sigma2)

    def forward(self, x1, x2, diag=False, **kwargs):
        Q = self.q * self.lambdas_sqrt
        v1 = x1 @ Q
        v2 = x2 @ Q

        if diag:
            min_size = min(x1.shape[0], x2.shape[0])
            kernel = self.sigma2 * torch.ones((min_size, ))
            
        else:
            # Equivalent to squared distance in the linear subspace
            z1 = torch.sum(torch.square(v1), axis=1).unsqueeze(1)
            z2 = torch.sum(torch.square(v2), axis=1).unsqueeze(0)
            kernel = torch.exp(self.log_sigma2 + 2 * v1 @ v2.T - z1 - z2)

        return kernel



# class SiteKernel(SequenceKernel):
#     def __init__(self, n_alleles, site, fixed_theta=False, **kwargs):
#         self.site = site
#         self.site_dims = torch.arange(site * n_alleles, (site + 1) * n_alleles)
#         self.fixed_theta = fixed_theta
#         super().__init__(n_alleles, seq_length=1, **kwargs)
#         self.set_params()

#     def select_site(self, x):
#         return(x.index_select(-1, self.site_dims.to(device=x.device)))

#     def calc_theta0(self):
#         if self.theta0 is not None:
#             theta0 = self.theta0
#         else:
#             theta0 = torch.full((self.theta_dim,), fill_value=self.theta_init)
#         return theta0

#     def set_params(self):
#         if self.fixed_theta:
#             self.theta = self.calc_theta0()
#         else:
#             theta0 = self.calc_theta0()
#             params = {"theta": Parameter(theta0, requires_grad=True)}
#             self.register_params(params)

#     def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
#         x1_, x2_ = self.select_site(x1), self.select_site(x2)
#         site_kernel = self.get_site_kernel()

#         if diag:
#             min_size = min(x1.shape[0], x2.shape[0])
#             kernel = torch.einsum("ia,ib,ab->i", x1_[:min_size], x2_[:min_size], site_kernel)
#         else:
#             kernel = torch.einsum("ia,jb,ab->ij", x1_, x2_, site_kernel)
#         return kernel

#     def _covar_func(self, x1, x2, **kwargs):
#         x1_ = LazyTensor(x1[..., :, None, :])
#         x2_ = LazyTensor(x2[..., None, :, :])
#         K = (x1_ * x2_).sum(-1)
#         return K

#     def _keops_forward(self, x1, x2, **kwargs):
#         site_kernel = self.get_site_kernel()
#         x1_ = self.select_site(x1) @ site_kernel
#         x2_ = self.select_site(x2)
#         kernel = KernelLinearOperator(x1_, x2_, covar_func=self._covar_func, **kwargs)
#         return kernel

#     def get_site_kernel(self):
#         return self.calc_site_kernel(self.theta)


# class ConnectednessSiteKernel(SiteKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, site, theta0=None, **kwargs):
#         self.theta0 = theta0
#         self.theta_init = -3.
#         self.theta_dim = 1
#         self.site_shape = (n_alleles, n_alleles)
#         super().__init__(n_alleles, site, **kwargs)


#     def calc_site_kernel(self, theta):
#         rho = torch.exp(theta)
#         v = (1 - rho) / (1 + (self.n_alleles - 1) * rho)
#         kernel = (torch.ones(self.site_shape, device=theta.device) * v).fill_diagonal_(1.0)
#         return(kernel)


# class JengaSiteKernel(SiteKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, site, theta0=None, **kwargs):
#         self.theta0 = theta0
#         self.theta_init = -3.
#         self.theta_dim = n_alleles + 1
#         super().__init__(n_alleles, site, **kwargs)

#     def calc_site_kernel(self, theta):
#         log_rho = theta[0].item()
#         rho = torch.exp(theta[0])
#         log_p = theta[1:] - torch.logsumexp(theta[1:], dim=0)
#         log_eta = log1mexp(log_p) - log_p
#         log_one_p_eta_rho = torch.logaddexp(torch.zeros_like(log_eta), log_rho + log_eta)
#         site_factors = torch.exp(-0.5 * log_one_p_eta_rho)
#         kernel = ((1 - rho) * site_factors.unsqueeze(0) * site_factors.unsqueeze(1)).fill_diagonal_(1.)
#         return(kernel)


# class GeneralSiteKernel(SiteKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, site, theta0=None, **kwargs):
#         self.theta0 = theta0
#         self.theta_init = -1.0
#         self.theta_dim = int(comb(n_alleles, 2))
#         super().__init__(n_alleles, site, **kwargs)
#         self.theta_to_L = CorrCholeskyTransform()

#     def calc_site_kernel(self, theta):
#         L = self.theta_to_L(theta)
#         site_kernel = L @ L.T
#         return site_kernel


# class SiteProductKernel(ProductKernel):
#     def __init__(self, n_alleles, seq_length, site_kernel, **kwargs):
#         kernels = [site_kernel(n_alleles, site, **kwargs) for site in range(seq_length)]
#         super().__init__(*kernels)


# class ConnectednessKernel(SiteProductKernel):
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length,
#                          site_kernel=ConnectednessSiteKernel, **kwargs)


# class JengaKernel(SiteProductKernel):
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length,
#                          site_kernel=JengaSiteKernel, **kwargs)


# class GeneralProductKernel(SiteProductKernel):
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length,
#                          site_kernel=GeneralSiteKernel, **kwargs)


# class RhoPiKernel(SequenceKernel):
#     def __init__(self, n_alleles, seq_length,
#                  logit_rho0=None, log_p0=None, log_var0=None,
#                  train_p=True, train_var=False,
#                  common_rho=False, correlation=False,
#                  random_init=False,
#                  **kwargs):
#         super().__init__(n_alleles, seq_length, **kwargs)
#         self.logit_rho0 = logit_rho0
#         self.random_init = random_init
#         self.log_p0 = log_p0
#         self.log_var0 = log_var0
#         self.train_p = train_p
#         self.train_var = train_var
#         self.correlation = correlation
#         self.common_rho = common_rho
#         self.set_params()

#     def get_log_p0(self):
#         if self.log_p0 is None:
#             log_p0 = -torch.ones((self.seq_length, self.n_alleles), dtype=self.dtype)
#         else:
#             log_p0 = self.log_p0
#         return(log_p0)

#     def get_logit_rho0(self):
#         # Choose rho0 so that correlation at l/2 is 0.1
#         if self.logit_rho0 is None:
#             shape = (1, 1) if self.common_rho else (self.seq_length, 1)
#             t = np.exp(-2 / self.seq_length * np.log(10.))
#             v = np.log((1 - t) / (self.n_alleles * t))
#             logit_rho0 = torch.full(shape, v, dtype=self.dtype) if self.logit_rho0 is None else self.logit_rho0
#             if self.random_init:
#                 logit_rho0 = torch.normal(logit_rho0, std=1.)
#         else:
#             logit_rho0 = self.logit_rho0
#         return(logit_rho0)

#     def get_log_var0(self, logit_rho0):
#         if self.log_var0 is None:
#             if self.correlation:
#                 rho = torch.exp(logit_rho0) / (1 + torch.exp(logit_rho0))
#                 log_var0 = torch.log(1 + (self.n_alleles - 1) * rho).sum()
#             else:
#                 log_var0 = torch.tensor(0, dtype=self.dtype)
#         else:
#             log_var0 = torch.tensor(self.log_var0, dtype=self.dtype)
#         return(log_var0)

#     def set_params(self):
#         logit_rho0 = self.get_logit_rho0()
#         log_p0 = self.get_log_p0()
#         log_var0 = self.get_log_var0(logit_rho0=logit_rho0)
#         params = {'logit_rho': Parameter(logit_rho0, requires_grad=True),
#                   'log_p': Parameter(log_p0, requires_grad=self.train_p),
#                   'log_var': Parameter(log_var0, requires_grad=self.train_var)}
#         self.register_params(params)

#     def get_log_eta(self):
#         log_p = self.log_p - torch.logsumexp(self.log_p, axis=1).unsqueeze(1)
#         log_eta = log1mexp(log_p) - log_p
#         return(log_eta)

#     def get_log_one_minus_rho(self):
#         return(-torch.logaddexp(self.zeros_like(self.logit_rho), self.logit_rho))

#     def get_factors(self):
#         log1mrho = self.get_log_one_minus_rho()
#         log_rho = self.logit_rho + log1mrho
#         log_eta = self.get_log_eta()
#         log_one_p_eta_rho = torch.logaddexp(self.zeros_like(log_rho), log_rho + log_eta)
#         factors = log_one_p_eta_rho - log1mrho

#         constant = log1mrho.sum()
#         if self.common_rho:
#             constant *= self.seq_length
#         constant += self.log_var
#         return(constant, factors, log_one_p_eta_rho)

#     def _nonkeops_forward(self, x1, x2, diag=False, **params):
#         constant, factors, log_one_p_eta_rho = self.get_factors()
#         factors = factors.reshape(1, self.t)
#         log_one_p_eta_rho = log_one_p_eta_rho.reshape(self.t, 1)

#         if diag:
#             min_size = min(x1.shape[0], x2.shape[0])
#             log_kernel = constant + (x1[:min_size, :] * x2[:min_size, :] * factors).sum(1)
#         else:
#             log_kernel = constant + x1 @ (x2 * factors).T

#         if self.correlation:
#             log_sd1 = 0.5 * (x1 @ log_one_p_eta_rho)
#             log_sd2 = 0.5 * (x2 @ log_one_p_eta_rho)
#             if diag:
#                 log_kernel = log_kernel - log_sd1.flatten() - log_sd2.flatten()
#             else:
#                 log_sd2 = log_sd2.reshape((1, x2.shape[0]))
#                 log_kernel = log_kernel - log_sd1 - log_sd2

#         kernel = torch.exp(log_kernel)
#         return(kernel)

#     def _covar_func(self, x1, x2, constant, **kwargs):
#         x1_ = LazyTensor(x1[..., :, None, :])
#         x2_ = LazyTensor(x2[..., None, :, :])
#         kernel = ((x1_ * x2_).sum(-1) + constant).exp()
#         return(kernel)

#     def _keops_forward(self, x1, x2, **kwargs):
#         # TODO: introduce constants before exponentiation in covar_func
#         constant, factors, log_one_p_eta_rho = self.get_factors()
#         f = factors.reshape(1, self.t)
#         kernel = KernelLinearOperator(x1, x2 * f,
#                                       covar_func=self._covar_func,
#                                       constant=constant, **kwargs)

#         if self.correlation:
#             log_one_p_eta_rho = log_one_p_eta_rho.reshape(1, self.t)
#             sd1_inv_D = DiagLinearOperator(torch.exp(-0.5 * (x1 * log_one_p_eta_rho).sum(1)))
#             sd2_inv_D = DiagLinearOperator(torch.exp(-0.5 * (x2 * log_one_p_eta_rho).sum(1)))
#             kernel = sd1_inv_D @ kernel @ sd2_inv_D

#         return(kernel)


# class ConnectednessKernel(RhoPiKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length,
#                          train_p=False, train_var=False,
#                          correlation=True,
#                          **kwargs)


#     def get_decay_rates(self, positions=None):
#         decay_rates = calc_decay_rates(self.logit_rho.detach().numpy(),
#                                        self.log_p.detach().numpy(),
#                                        sqrt=False, positions=positions).mean(1)
#         return(decay_rates)


# class JengaKernel(RhoPiKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length,
#                          correlation=True, train_p=True, train_var=False,
#                          **kwargs)

#     def get_decay_rates(self, alleles=None, positions=None):
#         decay_rates = calc_decay_rates(self.logit_rho.detach().numpy(),
#                                        self.log_p.detach().numpy(),
#                                        sqrt=True, alleles=alleles, positions=positions)
#         return(decay_rates)


# class ExponentialKernel(ConnectednessKernel):
#     def __init__(self, n_alleles, seq_length, **kwargs):
#         super().__init__(n_alleles, seq_length, common_rho=True, **kwargs)

#     def get_decay_rate(self):
#         decay_rate = calc_decay_rates(self.logit_rho.detach().numpy(),
#                                       self.log_p.detach().numpy(),
#                                       sqrt=False).values.mean()
#         return(decay_rate)


# class GeneralProductKernel_old(SequenceKernel):
#     is_stationary = True
#     def __init__(self, n_alleles, seq_length, theta0=None, **kwargs):
#         super().__init__(n_alleles, seq_length, **kwargs)
#         self.dim = int(comb(n_alleles, 2))
#         self.theta0 = theta0
#         self.set_params()
#         self.theta_to_L = CorrCholeskyTransform()

#     def calc_theta0(self):
#         if self.theta0 is not None:
#             theta0 = self.theta0
#         else:
#             theta0 = torch.zeros((self.seq_length, self.dim), dtype=self.dtype)
#         return(theta0)

#     def theta_to_covs(self, theta):
#         Ls = [self.theta_to_L(theta[i]) for i in range(self.seq_length)]
#         covs = torch.stack([(L @ L.T) for L in Ls], axis=0)
#         return(covs)

#     def set_params(self):
#         theta0 = self.calc_theta0()
#         params = {'theta': Parameter(theta0, requires_grad=True)}
#         self.register_params(params)

#     def get_covs(self):
#         return(self.theta_to_covs(self.theta))

#     def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
#         covs = self.get_covs()
#         K = self.inner_product(x1[:, :self.n_alleles], x2[:, :self.n_alleles],
#                                metric=covs[0, :, :], diag=diag)
#         for i in range(1, self.seq_length):
#             start, end = i * self.n_alleles, (i+1) * self.n_alleles
#             K *= x1[:, start:end] @ covs[i, :, :] @ x2[:, start:end].T
#         return(K)


# class GeneralProductKernel2(SequenceKernel):
#     is_stationary = True

#     def __init__(self, n_alleles, seq_length, theta0=None, **kwargs):
#         super().__init__(n_alleles, seq_length, **kwargs)
#         self.dim = int(comb(n_alleles, 2))
#         self.theta0 = theta0
#         self.set_params()

#     def calc_theta0(self):
#         if self.theta0 is not None:
#             theta0 = self.theta0
#         else:
#             theta0 = torch.zeros((self.seq_length, self.dim), dtype=self.dtype)
#         return theta0

#     def set_params(self):
#         theta0 = self.calc_theta0()
#         tril_indices = torch.tril_indices(row=self.n_alleles, col=self.n_alleles,
#                                           dtype=torch.int, offset=-1)
#         params = {"theta": Parameter(theta0, requires_grad=True),
#                   'idx': Parameter(tril_indices, requires_grad=False)}
#         self.register_params(params)

#     def theta_to_log_cov(self, theta):
#         v = -torch.exp(theta)
#         log_cov = torch.zeros((self.n_alleles, self.n_alleles),
#                               dtype=theta.dtype, device=theta.device)
#         log_cov[self.idx[0], self.idx[1]] = v
#         log_cov[self.idx[1], self.idx[0]] = v
#         return(log_cov)

#     def get_A(self):
#         return(torch.block_diag(*[self.theta_to_log_cov(self.theta[i, :])
#                                   for i in range(self.seq_length)]))

#     def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
#         A = self.get_A()
#         return(torch.exp(self.inner_product(x1, x2, metric=A, diag=diag)))

#     def _covar_func(self, x1, x2, **kwargs):
#         x1_ = LazyTensor(x1[:, None, :])
#         x2_ = LazyTensor(x2[None, :, :])
#         kernel = (x1_ * x2_).sum(-1).exp()
#         return(kernel)

#     def _keops_forward(self, x1, x2, **kwargs):
#         A = self.get_A()
#         kernel = KernelLinearOperator(x1, x2 @ A, covar_func=self._covar_func, **kwargs)
#         return(kernel)


# from gpytorch.lazy import delazify

# class AdditiveHeteroskedasticKernel(SequenceKernel):
#     @property
#     def is_stationary(self) -> bool:
#         return self.base_kernel.is_stationary

#     def __init__( self, base_kernel, n_alleles=None, seq_length=None,
#                   log_ds0=None, a=0.5, **kwargs):
#         if base_kernel.active_dims is not None:
#             kwargs["active_dims"] = base_kernel.active_dims

#         if hasattr(base_kernel, 'alpha'):
#             n_alleles = base_kernel.alpha
#         else:
#             if n_alleles is None:
#                 msg = 'If the base kernel does not have n_alleles attribute, '
#                 msg += 'it should be provided'
#                 raise ValueError(msg)

#         if hasattr(base_kernel, 'l'):
#             seq_length = base_kernel.l
#         else:
#             if seq_length is None:
#                 msg = 'If the base kernel does not have seq_length attribute, '
#                 msg += 'it should be provided'
#                 raise ValueError(msg)

#         super().__init__(n_alleles, seq_length, **kwargs)
#         self.log_ds0 = log_ds0
#         self.a = a
#         self.set_params()
#         self.base_kernel = base_kernel

#     def set_params(self):
#         theta = torch.zeros((self.seq_length, self.n_alleles)) if self.log_ds0 is None else self.log_ds0
#         params = {'theta': Parameter(theta, requires_grad=True),
#                   'theta0': Parameter(5 * torch.ones((1,)), requires_grad=True)}
#         self.register_params(params)

#     def get_theta(self):
#         t = self.theta
#         return(t - t.mean(1).unsqueeze(1))

#     def get_theta0(self):
#         return(self.theta0)

#     def f(self, x, theta0, theta, a=0, b=1):
#         phi = theta0 + (x * theta.reshape(1, 1, self.seq_length * self.n_alleles)).sum(-1)
#         r = a + (b - a) * torch.exp(phi) / (1 + torch.exp(phi))
#         return(r)

#     def forward(self, x1, x2, last_dim_is_batch=False, diag=False, **params):
#         orig_output = self.base_kernel.forward(x1, x2, diag=diag,
#                                                last_dim_is_batch=last_dim_is_batch,
#                                                **params)
#         theta0, theta = self.get_theta0(), self.get_theta()
#         f1 = self.f(x1, theta0, theta, a=self.a).T
#         f2 = self.f(x2, theta0, theta, a=self.a)

#         if last_dim_is_batch:
#             f1 = f1.unsqueeze(-1)
#             f2 = f2.unsqueeze(-1)
#         if diag:
#             f1 = f1.unsqueeze(-1)
#             f2 = f2.unsqueeze(-1)
#             return(f1 * f2 * delazify(orig_output))
#         else:
#             return(f1 * f2 * orig_output)

#     def num_outputs_per_input(self, x1, x2):
#         return self.base_kernel.num_outputs_per_input(x1, x2)

#     def prediction_strategy(self, train_inputs, train_prior_dist, train_labels, likelihood):
#         return self.base_kernel.prediction_strategy(train_inputs, train_prior_dist, train_labels, likelihood)



#!/usr/bin/env python
import unittest

import numpy as np
import torch
from scipy.special import comb

from epik.utils import (
    seq_to_one_hot,
    diploid_to_one_hot,
    get_full_space_one_hot,
    one_hot_to_seq,
    encode_seq,
    get_one_hot_subseq_key,
    get_binary_subseq_key,
    encode_seqs,
    calc_decay_rates,
    get_k_mutants,
    get_epistatic_coeffs_contrast_matrix,
    get_mut_effs_contrast_matrix,
    calc_distance_covariance,
    KrawtchoukPolynomials,
    SquaredMatMulOperator,
)


class UtilsTests(unittest.TestCase):
    def test_encode_seq(self):
        # Binary encoding
        seq = "ABBA"
        subseq_key = get_binary_subseq_key(alphabet="AB")
        assert len(subseq_key) == 2

        encoding = encode_seq(seq, subseq_key)
        assert encoding == [1, -1, -1, 1]
        assert len(subseq_key) == 5

        # Odd length
        seq = "ABBAB"
        encoding = encode_seq(seq, subseq_key)
        assert encoding == [1, -1, -1, 1, -1]
        assert len(subseq_key) > 5

        # Fail with missing characters
        seq = "ABBAC"
        try:
            encoding = encode_seq(seq, subseq_key)
            self.fail()
        except ValueError:
            pass

        # One hot encoding
        seq = "ACGT"
        subseq_key = get_one_hot_subseq_key(alphabet=seq)
        assert len(subseq_key) == 4

        encoding = encode_seq(seq, subseq_key)
        assert encoding == [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
        assert len(subseq_key) == 7

        encoding = encode_seq(seq[:3], subseq_key)
        assert encoding == [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0]

    def test_encode_seqs(self):
        seqs = np.array(["AA", "AB", "BA", "BB"])
        alphabet = "AB"

        onehot = np.array([[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1.0]])
        X = encode_seqs(seqs, alphabet, encoding_type="one_hot")
        assert np.allclose(X, onehot)

    def test_one_hot_encoding(self):
        X = np.array(["AA", "AB", "BA", "BB"])
        x = seq_to_one_hot(X).numpy()
        onehot = np.array([[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1.0]])
        assert np.allclose(x - onehot, 0)

    def test_one_hot_encoding_to_seq(self):
        x = np.array([[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1.0]])
        X = one_hot_to_seq(x, alleles=np.array(["A", "B"]))
        assert np.all(X == np.array(["AA", "AB", "BA", "BB"]))

    def test_get_full_one_hot(self):
        X = get_full_space_one_hot(seq_length=2, n_alleles=2)
        assert np.allclose(X, [[1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 0, 1], [0, 1, 0, 1]])

        X = get_full_space_one_hot(seq_length=2, n_alleles=3)
        assert np.allclose(
            X,
            [
                [1, 0, 0, 1, 0, 0],
                [0, 1, 0, 1, 0, 0],
                [0, 0, 1, 1, 0, 0],
                [1, 0, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0],
                [0, 0, 1, 0, 1, 0],
                [1, 0, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 1],
                [0, 0, 1, 0, 0, 1],
            ],
        )

    def test_diploid_encoding(self):
        X = np.array(["00", "01", "11", "02", "22"])
        x = diploid_to_one_hot(X).numpy()
        assert x.shape == (5, 2, 3)

        h0 = np.array([[1, 1], [1, 0], [0, 0], [1, 0], [0, 0]])
        ht = np.array([[0, 0], [0, 1], [1, 1], [0, 0], [0, 0]])
        h1 = np.array([[0, 0], [0, 0], [0, 0], [0, 1], [1, 1]])
        y = np.stack([h0, ht, h1], axis=2)
        assert np.allclose(x, y)

    def test_calc_decay_rates(self):
        logit_rho = np.array([[0.0], [-0.69], [0.69]])
        log_p = np.full((3, 3), 1 / 3.0)
        decay_rates = calc_decay_rates(
            logit_rho, log_p, alleles=["A", "B", "C"], positions=[10, 12, 15]
        )

        assert np.all(decay_rates.columns == ["A", "B", "C"])
        assert np.all(decay_rates.index == [10, 12, 15])

        rho = np.exp(logit_rho) / (1 + np.exp(logit_rho))
        expected_decay_rates = (1 - (1 - rho) / (1 + 2 * rho)).flatten()
        decay_rates = decay_rates.mean(1).values.flatten()
        assert np.allclose(decay_rates, expected_decay_rates)

        # With sqrt
        decay_rates = calc_decay_rates(
            logit_rho, log_p, sqrt=True, alleles=["A", "B", "C"], positions=[10, 12, 15]
        )
        expected_decay_rates = (1 - np.sqrt((1 - rho) / (1 + 2 * rho))).flatten()
        decay_rates = decay_rates.mean(1).values.flatten()
        assert np.allclose(decay_rates, expected_decay_rates)

    def test_get_k_mutants(self):
        seq0 = "AGCT"
        seqs = get_k_mutants(seq0, alleles=seq0, k=2)

        for seq in seqs:
            d = np.sum([a1 != a2 for a1, a2 in zip(seq, seq0)])
            assert d == 2

        expected_doubles = 3**2 * comb(4, 2)
        assert len(seqs) == expected_doubles

    def test_get_get_contrast_matrices(self):
        seq0 = "ACGT"

        contrast_matrix = get_mut_effs_contrast_matrix(seq0, alleles=seq0)
        n_contrasts, n_seqs = 12, 13
        assert contrast_matrix.shape == (n_contrasts, n_seqs)
        assert np.allclose(contrast_matrix.sum(1), 0)
        assert np.allclose((contrast_matrix != 0).sum(1), 2)

        contrast_matrix = get_epistatic_coeffs_contrast_matrix(seq0, alleles=seq0)
        n_contrasts = 3**2 * comb(4, 2)
        n_seqs = 1 + 3 * 4 + n_contrasts
        assert contrast_matrix.shape == (n_contrasts, n_seqs)
        assert np.allclose(contrast_matrix.sum(1), 0)
        assert np.allclose((contrast_matrix != 0).sum(1), 4)

    def test_calc_distance_covariance(self):
        seq_length = 2
        n = seq_length + 1
        x = np.tril(np.ones((seq_length + 1, seq_length)), k=-1)
        x = np.stack([x, 1 - x], axis=2).reshape(n, 2 * seq_length)
        x = torch.tensor(x, dtype=torch.float32)
        y = torch.normal(torch.zeros(x.shape[0]))

        _, ns = calc_distance_covariance(x, y, seq_length)
        assert ns[0] == 3.0
        assert ns[1] == 4
        assert ns[2] == 2

        # Verify chunks in longer sequences
        seq_length = 12
        n = seq_length + 1
        x = get_full_space_one_hot(seq_length=seq_length, n_alleles=2)
        y = torch.normal(torch.zeros(x.shape[0]))
        covs1, ns1 = calc_distance_covariance(x, y, seq_length, chunk_size=None)
        covs2, ns2 = calc_distance_covariance(x, y, seq_length, chunk_size=100)
        assert np.allclose(covs1[0], 1, atol=0.1)
        assert np.allclose(np.abs(covs1[1:]), 0, atol=0.1)
        assert np.allclose(covs1, covs2)
        assert np.allclose(ns1, ns2)

    def test_squared_matmul_operator(self):
        x1 = np.random.normal(size=(10, 3))
        x2 = np.random.normal(size=(10, 3))

        A = x1 @ x2.T
        target = A**2
        v = np.random.normal(size=10)
        u1 = target @ v

        x1, x2 = torch.Tensor(x1), torch.Tensor(x2)
        v = torch.Tensor(v)
        H2 = SquaredMatMulOperator(x1, x2.T)
        u2 = H2 @ v
        assert np.allclose(u1, u2, atol=1e-4)

    def test_krawtchouk_polynomials(self):
        ws = KrawtchoukPolynomials(n_alleles=2, seq_length=3)
        ns = ws.n_alleles ** ws.seq_length

        # Get constant covariance
        log_lambdas = torch.Tensor([0.0, -16, -16, -16])
        w_d = ws.get_w_d(log_lambdas)
        covs = torch.ones(4)
        assert torch.allclose(w_d, covs / ns)

        # Using the interpolating function
        w_d = ws.basis @ ws.get_c_b(log_lambdas)
        assert torch.allclose(w_d, covs / ns)

        w_d = ws.basis @ ws.c_bk @ torch.exp(log_lambdas)
        assert torch.allclose(w_d, covs / ns)

        # Get additive covariance
        log_lambdas = torch.Tensor([-16, 0.0, -16, -16])
        w_d = ws.get_w_d(log_lambdas)
        covs = torch.Tensor([3.0, 1, -1, -3])
        assert torch.allclose(w_d, covs / ns)

        # Using the polynomial in distance
        w_d = ws.basis @ ws.get_c_b(log_lambdas)
        assert torch.allclose(w_d, covs / ns)

        # Check in larger dimensions
        log_lambdas = torch.linspace(0, -6, 9)
        ws = KrawtchoukPolynomials(n_alleles=4, seq_length=8)
        w_d1 = ws.get_w_d(log_lambdas)
        w_d2 = ws.basis @ ws.get_c_b(log_lambdas)
        w_d3 = ws.basis @ ws.c_bk @ torch.exp(log_lambdas)
        assert torch.allclose(w_d1, w_d2, atol=1e-3)
        assert torch.allclose(w_d1, w_d3, atol=1e-3)

        log_lambdas = torch.linspace(0, -10, 5)
        ws = KrawtchoukPolynomials(n_alleles=20, seq_length=4)
        w_d1 = ws.get_w_d(log_lambdas)
        w_d2 = ws.basis @ ws.get_c_b(log_lambdas)
        w_d3 = ws.basis @ ws.c_bk @ torch.exp(log_lambdas)
        assert torch.allclose(w_d1, w_d2, atol=1e-3)
        assert torch.allclose(w_d1, w_d3, atol=1e-3)

        log_lambdas = torch.linspace(0, -10, 17)
        ws = KrawtchoukPolynomials(n_alleles=2, seq_length=16)
        w_d1 = ws.get_w_d(log_lambdas)
        w_d2 = ws.basis @ ws.get_c_b(log_lambdas)
        w_d3 = ws.basis @ ws.c_bk @ torch.exp(log_lambdas)
        assert torch.allclose(w_d1, w_d2, atol=1e-3)
        assert torch.allclose(w_d1, w_d3, atol=1e-3)


if __name__ == "__main__":
    import sys

    sys.argv = ["", "UtilsTests"]
    unittest.main()



#!/usr/bin/env python
import unittest

import numpy as np
import torch
from linear_operator import to_dense
from scipy.special import comb

from epik.kernel import (
    AdditiveKernel,
    ConnectednessKernel,
    ExponentialKernel,
    GeneralProductKernel,
    JengaKernel,
    PairwiseKernel,
    VarianceComponentKernel,
    FactorAnalysisKernel,
    SiteKernelAligner
)
from epik.utils import encode_seqs, get_full_space_one_hot


class KernelsTests(unittest.TestCase):
    def test_additive_kernel(self):
        sl, a = 1, 2
        x = get_full_space_one_hot(sl, a)
        Identity = torch.eye(a**sl)

        # Additive kernel with lambdas1 = 0 should return a constant matrix
        log_lambdas0 = torch.tensor([0.0, -10.0])
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, 1, atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (2,)
        assert np.allclose(cov, 1, atol=0.01)

        # Additive kernel with lambdas0 = 1 should return a purely additive cov
        log_lambdas0 = torch.tensor([-10.0, 0.0])
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, np.array([[1, -1], [-1, 1]]), atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (2,)
        assert np.allclose(cov, 1, atol=0.01)

        # Additive kernel with lambdas = 1 should return the identity
        log_lambdas0 = torch.tensor([0.0, 0])
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, (a**sl) * Identity, atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (2,)
        assert np.allclose(cov, a**sl, atol=0.01)

        sl, a = 2, 2
        x = get_full_space_one_hot(sl, a)

        # Constant kernel
        log_lambdas0 = torch.tensor([0.0, -10])
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, 1, atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (4,)
        assert np.allclose(cov, 1, atol=0.01)

        # Additive kernel
        log_lambdas0 = torch.tensor([-10.0, 0])
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [2, 0, 0, -2], atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (4,)
        assert np.allclose(cov, 2, atol=0.01)

        # Additive kernel with larger variance
        log_lambdas0 = torch.tensor([-10.0, np.log(2)]).to(dtype=torch.float32)
        kernel = AdditiveKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [4, 0, 0, -4], atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (4,)
        assert np.allclose(cov, 4, atol=0.01)

    def test_pairwise_kernel(self):
        sl, a = 2, 2
        x = get_full_space_one_hot(sl, a)

        # Constant kernel
        log_lambdas0 = torch.tensor([0.0, -10, -10])
        kernel = PairwiseKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, 1, atol=0.01)

        cov = kernel.forward(x, x, diag=True).detach().numpy()
        assert cov.shape == (4,)
        assert np.allclose(cov, 1, atol=0.01)

        # Additive kernel
        log_lambdas0 = torch.tensor([-10.0, 0, -10.0])
        kernel = PairwiseKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [2, 0, 0, -2], atol=0.01)

        # Additive kernel with larger variance
        log_lambdas0 = torch.tensor([-10.0, np.log(2), -10.0]).to(dtype=torch.float32)
        kernel = PairwiseKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [4, 0, 0, -4], atol=0.01)

        # Test in longer sequences
        np.random.seed(0)
        a, sl = 2, 16
        n = sl + 1
        x = np.tril(np.ones((sl + 1, sl)), k=-1)
        x = np.stack([x, 1 - x], axis=2).reshape(n, 2 * sl)
        x = torch.tensor(x, dtype=torch.float32)

        log_lambdas0 = torch.tensor([-30.0, 0, -30.0]).to(dtype=torch.float32)
        kernel = PairwiseKernel(n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0)
        cov1 = kernel.forward(x, x).detach().numpy()
        assert cov1.shape == (x.shape[0], x.shape[0])
        assert np.allclose(cov1, cov1.T)
        v = np.random.normal(size=cov1.shape[0])
        assert np.dot(v, cov1 @ v) > 0.0

    def test_vc_kernel(self):
        sl, a = 2, 2
        x = get_full_space_one_hot(sl, a)

        # k=0
        log_lambdas0 = torch.tensor([0, -20.0, -20.0], dtype=torch.float32)
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov, 1/4, atol=1e-4)

        diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
        assert np.allclose(diag, np.diag(cov))

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=1e-4)

        # k=1
        log_lambdas0 = torch.tensor([-10.0, 0.0, -10.0], dtype=torch.float32)
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        k1 = np.array(
            [[1, 0, 0, -1], [0, 1, -1, 0], [0, -1, 1, 0], [-1, 0, 0, 1]],
            dtype=np.float32,
        )
        assert np.allclose(cov, k1 / 2, atol=1e-4)

        diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
        assert np.allclose(diag, np.diag(cov))

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=1e-4)

        # k=2
        log_lambdas0 = torch.tensor([-20.0, -20.0, 0.0], dtype=torch.float32)
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        k2 = np.array(
            [[1, -1, -1, 1], [-1, 1, 1, -1], [-1, 1, 1, -1], [1, -1, -1, 1]],
            dtype=np.float32,
        )
        assert np.allclose(cov, k2 / 4., atol=1e-4)

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov)

        # With max k set
        log_lambdas0 = torch.tensor([0.0, -10])
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0, max_k=1
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov, 1 /4., atol=0.01)

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=0.01)

        # Additive kernel
        log_lambdas0 = torch.tensor([-20.0, 0])
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0, max_k=1
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [0.5, 0, 0, -0.5], atol=0.01)

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=0.01)

        # Additive kernel with larger variance
        log_lambdas0 = torch.tensor([-10.0, np.log(2)]).to(dtype=torch.float32)
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, log_lambdas0=log_lambdas0, max_k=1
        )
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0], [1, 0, 0, -1], atol=0.01)

        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=0.01)

        # Test in larger spaces
        a, sl, n = 2, 40, 3
        x = np.random.choice([0, 1], size=n * sl)
        x = np.vstack([x, 1 - x]).T.reshape((n, 2 * sl))
        x = torch.tensor(x, dtype=torch.float32)

        log_lambdas0 = torch.tensor([-10.0, 1, -30.0]).to(dtype=torch.float32)
        kernel = VarianceComponentKernel(
            n_alleles=a, seq_length=sl, max_k=2, log_lambdas0=log_lambdas0
        )
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
        assert np.allclose(cov2, cov, atol=1e-4)

        # Test PSD
        a, sl, n = 4, 8, 1000
        kernel = VarianceComponentKernel(n_alleles=a, seq_length=sl)

        for _ in range(10):
            seqs = ["".join(c) for c in np.random.choice(["A", "C", "G", "T"], size=(n, sl))]
            x = encode_seqs(seqs, alphabet='ACGT')
            K_xx = kernel(x, x)
            for _ in range(10):
                v = torch.rand(n)
                assert torch.dot(v, K_xx @ v) >= 0.0

    def test_exponential_kernel(self):
        sl, a = 2, 2
        theta0 = torch.Tensor(-np.log([2]))
        decay_factor = 1 / 3.0
        kernel = ExponentialKernel(n_alleles=a, seq_length=sl, theta0=theta0)

        x = get_full_space_one_hot(sl, a)
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov[0, :], [1, decay_factor, decay_factor, decay_factor**2])

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov, cov2)

        diag = kernel.forward(x, x, diag=True).detach().numpy()
        assert np.allclose(diag, np.diag(cov))

        # Check decay rate
        delta = kernel.get_delta().detach().numpy()
        assert(np.allclose(delta, 1 - decay_factor))
        
        # Check random initialization
        kernel = ExponentialKernel(n_alleles=a, seq_length=sl)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).detach().numpy()
        assert(np.allclose(cov1, cov2))

        # Check that it works for theta0 > 0
        theta0 = torch.Tensor(np.log([2]))
        decay_factor = -1 / 3.0
        kernel = ExponentialKernel(n_alleles=a, seq_length=sl, theta0=theta0)
        x = get_full_space_one_hot(sl, a)
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov[0, :], [1, decay_factor, decay_factor, decay_factor**2])

    def test_connectedness_kernel(self):
        sl, a = 1, 2
        log_rho = -np.log(2) * torch.ones(sl)
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl, theta0=log_rho)
        x = get_full_space_one_hot(sl, a)
        cov = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(cov[0, :], [1, 1 / 3.0])

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov)

        diag = kernel.forward(x, x, diag=True).detach().numpy()
        assert np.allclose(diag, np.diag(cov))
        
        # with 2 sites
        sl, a = 2, 2
        log_rho = -np.log(2) * torch.ones(sl)
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl, theta0=log_rho)
        x = get_full_space_one_hot(sl, a)
        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov[0, :], [1, 1 / 3.0, 1 / 3.0, 1 / 3.0**2])

        diag = kernel.forward(x, x, diag=True).detach().numpy()
        assert np.allclose(diag, np.diag(cov))

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov, cov)
        
        # With different variance
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl,
                                     theta0=log_rho,
                                     log_var0=torch.Tensor([-1.]))
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).detach().numpy()
        assert np.allclose(cov1, cov2)

        # With unequal decay factors
        log_rho = torch.tensor([-np.log(2), -np.log(3)], dtype=torch.float32)
        rho = torch.exp(log_rho)
        decay_factors = (1 - rho) / (1 + rho)

        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl, theta0=log_rho)

        cov = kernel.forward(x, x).detach().numpy()
        expected = np.array(
            [1, decay_factors[0], decay_factors[1], decay_factors[0] * decay_factors[1]]
        )
        assert np.allclose(cov[0, :], expected)

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov)
        
        # Check decay rates calculation
        delta = kernel.get_delta().detach().numpy()
        expected_delta = 1 - decay_factors
        assert(np.allclose(delta, expected_delta))
        
        # Check random initialization
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).detach().numpy()
        assert(np.allclose(cov1, cov2))
        
        # Check longer sequences
        sl, a = 6, 4
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl)
        x = get_full_space_one_hot(sl, a)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov1)
        
        # Check longer sequences
        kernel = ConnectednessKernel(n_alleles=a, seq_length=sl,
                                     log_var0=torch.Tensor([-1.]))
        x = get_full_space_one_hot(sl, a)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov1)

    def test_jenga_kernel(self):
        sl, a = 1, 3
        theta0 = torch.tensor(np.log([[0.5, 0.2, 0.6, 0.2]]), dtype=torch.float32)
        kernel = JengaKernel(n_alleles=a, seq_length=sl, theta0=theta0)
        x = get_full_space_one_hot(sl, a)
        rho, eta = 0.5, np.array([4.0, 2 / 3, 4.0])
        allele_factors = np.sqrt(1 + rho * eta)
        a01, a02, a12 = (
            allele_factors[0] * allele_factors[1],
            allele_factors[0] * allele_factors[2],
            allele_factors[1] * allele_factors[2],
        )
        expected = np.array(
            [
                [1, (1 - rho) / a01, (1 - rho) / a02],
                [(1 - rho) / a01, 1, (1 - rho) / a12],
                [(1 - rho) / a02, (1 - rho) / a12, 1],
            ]
        )

        cov = kernel.forward(x, x).detach().numpy()
        assert np.allclose(cov, expected)

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov)

        # With two sites
        sl, a = 2, 2
        theta0 = torch.tensor(
            np.log([[0.5, 0.2, 0.8], [0.5, 0.5, 0.5]]), dtype=torch.float32
        )
        kernel = JengaKernel(n_alleles=a, seq_length=sl, theta0=theta0)
        x = get_full_space_one_hot(sl, a)
        rho = np.array([0.5, 0.5])
        eta = np.array([[4.0, 0.25], [1.0, 1.0]])
        cov = kernel.forward(x, x).detach().numpy()
        expected = np.array(
            [
                1,
                (1 - rho[0])
                / np.sqrt((1 + rho[0] * eta[0, 0]) * (1 + rho[0] * eta[0, 1])),
                (1 - rho[1])
                / np.sqrt((1 + rho[1] * eta[1, 0]) * (1 + rho[1] * eta[1, 1])),
                np.nan,
            ]
        )
        expected[3] = expected[1] * expected[2]
        assert np.allclose(cov[0, :], expected)

        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov)

        # Check decay rates
        rho = np.expand_dims(rho, 1)
        delta = kernel.get_delta().detach().numpy()
        expected_delta = 1 - np.sqrt((1-rho) / (1 + eta * rho))
        assert(np.allclose(delta, expected_delta))
        
        # Check random initialization
        kernel = JengaKernel(n_alleles=a, seq_length=sl)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).detach().numpy()
        assert(np.allclose(cov1, cov2))
        
        # Check longer sequences
        sl, a = 6, 4
        kernel = JengaKernel(n_alleles=a, seq_length=sl)
        x = get_full_space_one_hot(sl, a)
        cov1 = kernel._nonkeops_forward(x, x).detach().numpy()
        cov2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(cov2, cov1)
        
    def test_general_product_kernel(self):
        sl, a = 2, 2
        x = get_full_space_one_hot(sl, a)

        theta0 = torch.full((sl, 1), fill_value=0.0)
        kernel = GeneralProductKernel(a, sl, theta0=theta0)
        K = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.allclose(K, np.eye(a**sl))

        K = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(K, np.eye(a**sl))
        
        # Check decay factors
        delta = kernel.get_delta().detach().numpy()
        assert(np.allclose(delta, 1 - np.eye(a)))

        # Check non HOC model
        theta0 = torch.full((sl, 1), fill_value=-1.0)
        kernel = GeneralProductKernel(a, sl, theta0=theta0)
        K1 = kernel._nonkeops_forward(x, x).detach().numpy()
        assert np.any(K1 != np.eye(a**sl))

        K2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(K1, K2)
        
        # Random initialization
        kernel = GeneralProductKernel(a, sl)
        K1 = kernel._nonkeops_forward(x, x).detach().numpy()
        K2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(K1, K2)

        # With larger spaces and random values
        seqs = ["ACGTAGCTAA", "GGGTAGCTAA", "GGGTAGCTCC"]
        x = encode_seqs(seqs, alphabet="ACGT")
        a, sl = 4, 10
        ncomb = int(comb(a, 2))
        theta0 = torch.normal(torch.zeros(size=(sl, ncomb)))
        kernel = GeneralProductKernel(a, sl, theta0=theta0)
        K1 = kernel._nonkeops_forward(x, x).detach().numpy()
        K2 = kernel._keops_forward(x, x).to_dense().detach().numpy()
        assert np.allclose(K1, K2)
        assert np.allclose(np.diag(K1), 1.0)
    
    def test_factor_analysis_kernel(self):
        sl, a, k = 4, 4, 3
        x = get_full_space_one_hot(sl, a)

        kernel = FactorAnalysisKernel(a, sl, k)
        K = kernel.forward(x, x).detach().numpy()
        assert(np.allclose(np.diag(K), 1.0))
        assert(np.allclose(K, K.T))

        # Ensure PSD
        for _ in range(10):
            v = np.random.normal(size=a**sl)
            assert(np.dot(v, K @ v) >= 0.)

        K = kernel.forward(x, x, diag=True).detach().numpy()
        assert np.allclose(K, 1.0)
        
    def test_kernel_aligner(self):
        n_alleles, seq_length = 2, 2
        aligner = SiteKernelAligner(n_alleles=n_alleles, 
                                    seq_length=seq_length)
        
        # Test Exponential transforms
        theta1 = np.array([[-1]])
        
        theta2 = aligner.exponential_to_connectedness(theta1)
        assert(theta2.shape == (2, 1))
        assert(np.allclose(theta2[0], -1))
        assert(np.allclose(aligner.connectedness_to_exponential(theta2), theta1))
        
        theta3 = aligner.exponential_to_jenga(theta1)
        assert(np.allclose(theta3[:, [0]], -1))
        assert(np.allclose(theta3[:, 1:], 0.))
        assert(np.allclose(aligner.jenga_to_exponential(theta3), theta1))
        
        theta4 = aligner.exponential_to_general_product(theta1)
        assert(np.allclose(aligner.general_product_to_exponential(theta4), theta1))
        
        # Test Connectedness transforms
        theta1 = np.array([[-1],
                           [-2]])
        
        theta2 = aligner.connectedness_to_exponential(theta1)
        assert(theta2.shape == (1, 1))
        assert(np.allclose(aligner.log_rho_to_q(theta2[0])[0], 0.61185566))
        
        theta3 = aligner.connectedness_to_jenga(theta1)
        assert(theta3.shape == (seq_length, n_alleles + 1))
        assert(np.allclose(theta3[:, [0]], theta1))
        assert(np.allclose(theta3[:, 1:], 0.))
        assert(np.allclose(aligner.jenga_to_connectedness(theta3), theta1))
        
        theta4 = aligner.connectedness_to_general_product(theta1)
        assert(np.allclose(aligner.general_product_to_connectedness(theta4), theta1))
        
        # Test Jenga transform
        n_alleles, seq_length = 4, 4
        theta1 = np.random.normal(size=(seq_length, n_alleles + 1))
        aligner = SiteKernelAligner(n_alleles=n_alleles, 
                                    seq_length=seq_length)
        
        theta2 = aligner.jenga_to_general_product(theta1)
        theta3 = aligner.general_product_to_jenga(theta2)
        for i in range(seq_length):
            c1 = aligner.jenga_to_corr(theta1[i])
            c3 = aligner.jenga_to_corr(theta3[i])
            assert(np.allclose(c1, c3, atol=1e-2))
        

    # def test_connectedness_site_kernel(self):
    #     sl, a = 2, 2
    #     x = get_full_space_one_hot(sl, a)
    #     theta0 = torch.full((1,), fill_value=np.log(0.5))
    #     rho0 = torch.exp(theta0)[0].item()
    #     r0 = (1 - rho0) / (1 + (a - 1) * rho0)
    #     k0 = np.ones((2, 2))
    #     k1 = np.array([[1, r0], [r0, 1]])
    #     K1 = np.kron(k0, k1)
    #     K2 = np.kron(k1, k0)

    #     # Site 1 kernel
    #     kernel = ConnectednessSiteKernel(a, site=0, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K1)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K1)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.0)

    #     # Site 2 kernel
    #     kernel = ConnectednessSiteKernel(a, site=1, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K2)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K2)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.0)

    # def test_jenga_site_kernel(self):
    #     sl, a = 2, 3
    #     x = get_full_space_one_hot(sl, a)
    #     theta0 = torch.Tensor(-np.log([2, 2, 4, 4]))
    #     r1 = 0.5 / (np.sqrt(2.5) * np.sqrt(1.5))
    #     r2 = 1 / 5
    #     k0 = np.ones((3, 3))
    #     k1 = np.array([[1, r1, r1], [r1, 1, r2], [r1, r2, 1]])
    #     K1 = np.kron(k0, k1)
    #     K2 = np.kron(k1, k0)

    #     # Site 1 kernel
    #     kernel = JengaSiteKernel(a, site=0, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K1)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K1)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.0)

    #     # Site 2 kernel
    #     kernel = JengaSiteKernel(a, site=1, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K2)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K2)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.0)

    # def test_general_site_kernel(self):
    #     sl, a = 2, 2
    #     x = get_full_space_one_hot(sl, a)
    #     K1 = np.array([[1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1]])
    #     K2 = np.array([[1, 1, 0, 0], [1, 1, 0, 0], [0, 0, 1, 1], [0, 0, 1, 1]])

    #     theta0 = torch.full((1,), fill_value=0.0)

    #     # Site 1 kernel
    #     kernel = GeneralSiteKernel(a, site=0, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K1)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K1)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.)

    #     # Site 2 kernel
    #     kernel = GeneralSiteKernel(a, site=1, theta0=theta0)
    #     K = kernel._nonkeops_forward(x, x).detach().numpy()
    #     assert np.allclose(K, K2)

    #     K = kernel._keops_forward(x, x).to_dense().detach().numpy()
    #     assert np.allclose(K, K2)

    #     K_diag = kernel._nonkeops_forward(x, x, diag=True).detach().numpy()
    #     assert K_diag.shape == (x.shape[0],)
    #     assert np.allclose(K_diag, 1.0)

    # def test_rho_pi_kernel(self):
    #     sl, a = 1, 2
    #     logit_rho0 = torch.tensor([[0.]])
    #     log_p0 = torch.tensor(np.log([[0.2, 0.8]]), dtype=torch.float32)
    #     kernel = RhoPiKernel(n_alleles=a, seq_length=sl,
    #                          logit_rho0=logit_rho0, log_p0=log_p0)
    #     x = get_full_space_one_hot(sl, a)
    #     cov = kernel.forward(x, x).detach().numpy()
    #     expected = np.array([[3, 0.5],
    #                          [0.5, 1.125]])
    #     assert(np.allclose(cov, expected))

    #     diag = kernel.forward(x, x, diag=True).detach().numpy()
    #     assert(np.allclose(diag, np.diag(cov)))

    #     cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
    #     assert(np.allclose(cov2, cov))

    #     sl, a = 2, 2
    #     logit_rho0 = torch.tensor([[0.],
    #                                [0.]], dtype=torch.float32)
    #     log_p0 = torch.tensor(np.log([[0.2, 0.8],
    #                                   [0.5, 0.5]]), dtype=torch.float32)
    #     kernel = RhoPiKernel(n_alleles=a, seq_length=sl,
    #                          logit_rho0=logit_rho0, log_p0=log_p0)
    #     x = get_full_space_one_hot(sl, a)
    #     rho = np.array([0.5, 0.5])
    #     eta = np.array([[4., 0.25],
    #                     [1., 1.  ]])
    #     cov = kernel.forward(x, x).detach().numpy()
    #     expected = np.array([(1 + rho[0] * eta[0, 0]) * (1 + rho[1] * eta[1, 0]),
    #                          (1 - rho[0])             * (1 + rho[1] * eta[1, 0]),
    #                          (1 + rho[0] * eta[0, 0]) * (1 - rho[1]),
    #                          (1 - rho[0])             * (1 - rho[1])])
    #     assert(np.allclose(cov[0, :], expected))

    #     diag = kernel.forward(x, x, diag=True).detach().numpy()
    #     assert(np.allclose(diag, np.diag(cov)))

    #     cov2 = to_dense(kernel._keops_forward(x, x)).detach().numpy()
    #     assert(np.allclose(cov2, cov))

    # def test_heteroskedastic_kernel(self):
    #     sl, a = 1, 2
    #     x = get_full_space_one_hot(sl, a)

    #     logit_rho0 = torch.tensor([[0.0]])
    #     kernel = ConnectednessKernel(n_alleles=a, seq_length=sl, logit_rho0=logit_rho0)
    #     cov1 = kernel.forward(x, x)
    #     assert cov1[0, 0] == 1.5
    #     assert cov1[0, 1] == 0.5

    #     kernel = AdditiveHeteroskedasticKernel(kernel)
    #     cov2 = kernel.forward(x, x)
    #     assert cov2[0, 0] < 1.5
    #     assert cov2[0, 1] < 0.5

    # def test_keops(self):
    #     from pykeops.torch import LazyTensor

    #     # Example inputs x1 and x2, shapes: (batch_size, n, d)
    #     x1 = torch.randn(10, 100, 10)  # Shape: (batch_size, n, d)
    #     x2 = torch.randn(10, 100, 10)  # Shape: (batch_size, n, d)

    #     # Step 1: Convert x1 and x2 into LazyTensor objects
    #     x1_lazy = LazyTensor(x1[:, None, :, :])  # Shape: (batch_size, n, 1, d)
    #     x2_lazy = LazyTensor(x2[:, :, None, :])  # Shape: (batch_size, n, d, 1)

    #     # Step 2: Compute the element-wise product over the last axis
    #     elementwise_product = x1_lazy * x2_lazy  # Shape: (batch_size, n, d, d)

    #     # Step 3: Perform sum reduction over axis `a` (axis=-2 in this case)
    #     summed_result = elementwise_product.sum(-1)  # Shape: (batch_size, n, d)
    #     print(summed_result.shape)

    #     # Step 4: Perform product reduction over axis `l` (axis=-1 in this case)
    #     final_result = summed_result.sum(-3)  # Shape: (batch_size, n)

    #     print(final_result.shape)


if __name__ == "__main__":
    import sys

    sys.argv = ["", "KernelsTests"]
    unittest.main()



import sys
import time
from collections import defaultdict
from itertools import combinations, product

import numpy as np
import pandas as pd
import torch
from linear_operator.operators import LinearOperator, MatmulLinearOperator
from scipy.special import logsumexp


def get_random_sequences(n, seq_length, alphabet):
    alleles = np.random.choice(alphabet, size=(n, seq_length), replace=True)
    seqs = np.array([''.join(s) for s in alleles])
    return(seqs)


def get_one_hot_subseq_key(alphabet, max_l=1):
    subseq_key = {}
    for i, c in enumerate(alphabet):
        z = [0] * len(alphabet)
        z[i] = 1
        subseq_key[c] = z

    if max_l > 1:
        for k in range(2, max_l):
            for alleles in product(alphabet, repeat=k):
                seq = "".join(alleles)
                subseq_key[seq] = []
                for c in alleles:
                    subseq_key[seq] += subseq_key[c]
    return subseq_key


def get_binary_subseq_key(alphabet):
    if len(alphabet) != 2:
        raise ValueError("Alphabet length must be 2")

    subseq_key = {alphabet[0]: [1], alphabet[1]: [-1]}
    return subseq_key


def encode_seq(seq, subseq_key, max_l=4):
    try:
        return subseq_key[seq]
    except KeyError:
        sl = len(seq)
        if sl == 1:
            raise ValueError("Missing characters in `subseq_key`")

        i = sl // 2
        enc1 = encode_seq(seq[:i], subseq_key=subseq_key)
        enc2 = encode_seq(seq[i:], subseq_key=subseq_key)
        encoding = enc1 + enc2

        if sl <= max_l:
            subseq_key[seq] = encoding

        return encoding


def get_alleles(c, alleles=None):
    if alleles is not None:
        return alleles
    else:
        return np.unique(c)


def encode_seqs(seqs, alphabet, encoding_type="one_hot", max_n=500):
    """
    Returns a torch.Tensor with the encoding of the provided sequences

    Parameters
    ----------
    seqs : array-like of shape (n_sequences,)
        Array containing a list of sequences
    alphabet : array-like of shape (n_alleles, )
        Iterable object with the list of alleles in the sequences
    encoding_type : 'one_hot' or 'binary'
        Type of encoding to use. If `encoding_type='one_hot'`,
        sequences will be encoded under the one-hot encoding
        with one column for every allele and site.
        If `encoding_type='binary'` and there are only two possible
        alleles, then sequences will be encoded as binary, with a
        single column for every site taking values (-1, 1)
    max_n : int (500)
        Maximum number of features to allow in the encoding

    Returns
    -------
    X : torch.Tensor of shape (n_sequences, n_features)
        Tensor with the numerical encoding of the input sequences
    """

    max_l = max_n // len(alphabet)
    if encoding_type == "one_hot":
        subseq_key = get_one_hot_subseq_key(alphabet)
    elif encoding_type == "binary":
        subseq_key = get_binary_subseq_key(alphabet)
    else:
        raise ValueError("encoding_type can only be `one_hot` or `binary`")

    X = get_tensor(
        [encode_seq(seq, subseq_key=subseq_key, max_l=max_l) for seq in seqs]
    )
    return X


def seq_to_one_hot(X, alleles=None):
    m = np.array([[a for a in x] for x in X])
    onehot = []
    for i in range(m.shape[1]):
        c = m[:, i]
        for allele in get_alleles(c, alleles=alleles):
            onehot.append(get_tensor(c == allele))
    onehot = torch.stack(onehot, 1).contiguous()
    return onehot


def seq_to_binary(X, ref):
    m = np.array([[a for a in x] for x in X])
    onehot = []
    for i in range(m.shape[1]):
        c = m[:, i]
        onehot.append(2 * get_tensor(c == ref) - 1)
    onehot = torch.stack(onehot, 1).contiguous()
    return onehot


def one_hot_to_seq(x, alleles):
    ncol = x.shape[1]
    alpha = alleles.shape[0]
    if ncol % alpha != 0:
        raise ValueError("Check that the number of alleles is correct")
    seq_length = int(ncol / alpha)
    a = np.hstack([alleles] * seq_length)
    X = ["".join(a[i == 1]) for i in x]
    return X


def diploid_to_one_hot(X, dtype=torch.float32):
    m = torch.tensor([[int(a) for a in x] for x in X])
    onehot = torch.stack([m == 0, m == 1, m == 2], 2).to(dtype=dtype)
    return onehot


def to_device(tensor, output_device=None):
    if output_device is not None and output_device != -1:
        tensor = tensor.to(output_device)
    return tensor


def get_tensor(ndarray, dtype=torch.float32, device=None):
    if not torch.is_tensor(ndarray):
        ndarray = torch.tensor(ndarray, dtype=dtype)
    if ndarray.dtype != dtype:
        ndarray = ndarray.to(dtype=dtype)
    return to_device(ndarray, output_device=device)


def to_numpy(v):
    u = v.detach()
    if u.is_cuda:
        u = u.cpu()
    return(u.numpy())


def get_gpu_memory(device=None):
    mem = torch.cuda.memory_allocated(device) / 2**20
    suffix = "MB"

    if mem < 1:
        mem = mem * 2**10
        suffix = "KB"

    return "{:.0f}{}".format(mem, suffix)


class LogTrack(object):
    def __init__(self, fhand=None):
        if fhand is None:
            fhand = sys.stderr
        self.fhand = fhand
        self.start = time.time()

    def write(self, msg, add_time=True):
        if add_time:
            msg = "[ {} ] {}\n".format(time.ctime(), msg)
        else:
            msg += "\n"
        self.fhand.write(msg)
        self.fhand.flush()

    def finish(self):
        t = time.time() - self.start
        self.write("Finished succesfully. Time elapsed: {:.1f} s".format(t))


def guess_space_configuration(seqs):
    """
    Guess the sequence space configuration from a collection of sequences
    This allows to have different number of alleles per site and maintain
    the order in which alleles appear in the sequences when enumerating the
    alleles per position

    Parameters
    ----------
    seqs: array-like of shape (n_genotypes,)
        Vector or list containing the sequences from which we want to infer
        the space configuration


    Returns
    -------
    config: dict with keys {'seq_length', 'n_alleles', 'alphabet'}
            Returns a dictionary with the inferred configuration of the discrete
            space where the sequences come from.

    """

    alleles = defaultdict(dict)
    for seq in seqs:
        for i, a in enumerate(seq):
            alleles[i][a] = 1
    length = len(alleles)
    config = {
        "seq_length": length,
        "n_alleles": [len(alleles[i]) for i in range(length)],
        "alphabet": [[a for a in alleles[i].keys()] for i in range(length)],
    }
    return config


def split_training_test(X, y, y_var=None, ptrain=0.8, dtype=None):
    """
    Splits a dataset into training and test subsets

    Parameters
    ----------
    X : torch.Tensor of shape (n_sequences, n_features)
        Tensor with the numerical encoding of the input sequences
    y : torch.Tensor of shape (n_sequence,)
            Tensor containing the phenotypic measurements for each
            sequence in `X`
    y_var : torch.Tensor of shape (n_sequence,) or None
        If `y_var=None` it is assumed that there is no uncertainty
        in the measurements. Otherwise, Tensor containing the
        variance of the measurements in `y`.
    ptrain : float (0.8)
        Proportion of the dataset to keep as training set
    dtype : torch.dtype (torch.float32)
        dtype to use for storing the output tensors

    Returns
    -------
    output: tuple of size (5,)
        Tuple containing (train_X, train_y, test_X, test_y, test_y_var),
        where `X` is the matrix encoding the input sequences,
        `y` the associated measurements and `y_var` the variance of
        those measurements.
    """

    ps = np.random.uniform(size=X.shape[0])
    train = ps <= ptrain
    train_x, train_y = X[train, :], y[train]

    if y_var is None:
        train_y_var = None
    else:
        train_y_var = y_var[train]

    test = ps > ptrain
    test_x, test_y = X[test, :], y[test]

    output = [train_x, train_y, test_x, test_y, train_y_var]
    if dtype is not None:
        output = [get_tensor(a, dtype=dtype) if a is not None else None for a in output]
    return output


def ps_to_variances(ps):
    v = 1 / ps
    v = (v.T / v.sum(1)).T
    return v


def get_full_space_one_hot(seq_length, n_alleles, dtype=torch.float32):
    n = n_alleles**seq_length
    i = torch.arange(n)

    c = i
    one_hot = []
    for _ in range(seq_length):
        r = c % n_alleles
        for j in range(n_alleles):
            one_hot.append(r == j)
        c = torch.div(c, n_alleles, rounding_mode="floor")
    X = torch.vstack(one_hot).T.to(dtype=dtype).contiguous()
    return X


def get_full_space_binary(seq_length, dtype=torch.float32):
    if seq_length == 1:
        return torch.tensor([[1], [-1]], dtype=dtype)
    else:
        b1 = get_full_space_binary(seq_length - 1, dtype=dtype)
        ones = torch.ones((b1.shape[0], 1), dtype=dtype)

        return torch.vstack([torch.hstack([b1, ones]), torch.hstack([b1, -ones])])


def log1mexp(x):
    """Numerically accurate evaluation of log(1 - exp(x)) for x < 0.
    See [Maechler2012accurate]_ for details.
    """
    two = torch.tensor([2.0], dtype=x.dtype).to(device=x.device)
    mask = -torch.log(two) < x  # x < 0
    return torch.where(mask, (-x.expm1()).log(), (-x.exp()).log1p())


def calc_vandermonde_inverse(values, n=None):
    s = values.shape[0]
    if n is None:
        n = s

    B = np.zeros((n, s))
    idx = np.arange(s)
    sign = np.zeros((n, s))

    for k in range(s):
        v_k = values[idx != k]
        norm_factor = 1 / np.prod(v_k - values[k])

        for power in range(n):
            v_k_combs = list(combinations(v_k, s - power - 1))
            p = np.sum([np.prod(v) for v in v_k_combs])
            B[power, k] = sign[power, k] * p * norm_factor
            sign[power, k] = (-1) ** (power)
            print(1 / norm_factor, p)
    return B


def log_factorial(x):
    return torch.lgamma(x + 1)


def log_comb(n, k):
    return log_factorial(n) - log_factorial(k) - log_factorial(n - k)


def calc_decay_rates(logit_rho, log_p, sqrt=False, alleles=None, positions=None):
    rho = np.exp(logit_rho) / (1 + np.exp(logit_rho))
    p = np.exp(log_p - np.expand_dims(logsumexp(log_p, axis=1), 1))
    eta = (1 - p) / p

    decay_factors = (1 - rho) / (1 + eta * rho)
    if sqrt:
        decay_factors = np.sqrt(decay_factors)

    decay_rates = pd.DataFrame(1 - decay_factors, index=positions, columns=alleles)
    return decay_rates


def get_k_mutants(seq0, alleles, k=2):
    positions = np.arange(len(seq0))
    seqs = []
    for sites in combinations(positions, k):
        for new_alleles in product(alleles, repeat=k):
            new_seq = [c for c in seq0]
            skip = False
            for a, s in zip(new_alleles, sites):
                if a == seq0[s]:
                    skip = True
                    break
                new_seq[s] = a
            if not skip:
                seqs.append("".join(new_seq))
    return seqs


def get_mutant_seq(seq, sites, alleles):
    mut = [c for c in seq]
    for p, a in zip(sites, alleles):
        mut[p] = a
    return "".join(mut)


def get_epistatic_coeffs_contrast_matrix(seq0, alleles):
    positions = np.arange(len(seq0))
    contrasts = {}
    for s1, s2 in combinations(positions, 2):
        c1, c2 = seq0[s1], seq0[s2]
        alleles1 = [a for a in alleles if a != c1]
        alleles2 = [a for a in alleles if a != c2]

        for a1 in alleles1:
            for a2 in alleles2:
                seqs = [
                    seq0,
                    get_mutant_seq(seq0, sites=[s1], alleles=[a1]),
                    get_mutant_seq(seq0, sites=[s2], alleles=[a2]),
                    get_mutant_seq(seq0, sites=[s1, s2], alleles=[a1, a2]),
                ]
                label = "{}{}{}_{}{}{}".format(c1, s1, a1, c2, s2, a2)
                values = [1, -1, -1, 1]
                contrasts[label] = dict(zip(seqs, values))
    contrasts_matrix = pd.DataFrame(contrasts).fillna(0).T
    return contrasts_matrix


def get_mut_effs_contrast_matrix(seq0, alleles):
    positions = np.arange(len(seq0))
    contrasts = {}
    for s in positions:
        c = seq0[s]
        for a in [a for a in alleles if a != c]:
            seqs = [seq0, get_mutant_seq(seq0, sites=[s], alleles=[a])]
            label = "{}{}{}".format(c, s, a)
            values = [-1, 1]
            contrasts[label] = dict(zip(seqs, values))
    contrasts_matrix = pd.DataFrame(contrasts).fillna(0).T
    return contrasts_matrix


def calc_distance_covariance(X, y, seq_length, chunk_size=None):
    if chunk_size is None:
        chunk_size = X.shape[0]

    cov = torch.zeros(seq_length + 1)
    ns = torch.zeros(seq_length + 1)

    nchunks = int(X.shape[0] / chunk_size) + 1
    ones = torch.ones_like(y)

    for i in range(nchunks):
        start, end = i * chunk_size, (i + 1) * chunk_size
        s = X[start:end, :] @ X.T
        chunk_y = y[start:end]
        chunk_ones = torch.ones_like(chunk_y)

        for v in range(seq_length + 1):
            d = int(seq_length - v)
            A = (s == v).to(dtype=torch.float)
            ns_d = torch.dot(chunk_ones, A @ ones).item()
            ns[d] += ns_d
            cov[d] += torch.dot(chunk_y, A @ y).item()
    cov = cov / ns
    cov[torch.isnan(cov)] = 0.0
    return (cov, ns)


class KrawtchoukPolynomials(object):
    def __init__(self, n_alleles, seq_length, max_k=None):
        if max_k is None or max_k > seq_length:
            max_k = seq_length

        self.n_alleles = n_alleles
        self.seq_length = seq_length
        self.max_k = max_k

        self.calc_log_w_dkq()
        self.calc_c_bk()

    def calc_log_w_dkq(self):
        size1 = self.seq_length + 1
        size2 = self.max_k + 1
        log_a_minus_1 = np.log(self.n_alleles - 1)
        l_log_a = self.seq_length * np.log(self.n_alleles)
        d = torch.arange(size1).reshape((size1, 1, 1))
        k = torch.arange(size2).reshape((1, size2, 1))
        q = torch.arange(size2).reshape((1, 1, size2))

        self.log_w_dkq = (
            (k - q) * log_a_minus_1
            + log_comb(d, q)
            + log_comb(self.seq_length - d, k - q)
            - l_log_a
        )
        self.w_dkq_sign = (-1.0) ** q
        self.w_dk = (self.w_dkq_sign * torch.exp(self.log_w_dkq)).sum(-1)

    def calc_w_d(self, w_dkq_sign, log_w_dkq, log_lambdas):
        w_dkq = torch.exp(log_w_dkq + log_lambdas[None, :, None])
        return (w_dkq_sign * w_dkq).sum((1, 2))

    def calc_c_bk(self):
        theta = -np.log(0.1) / self.seq_length
        x = torch.arange(self.seq_length + 1).to(dtype=torch.float)
        self.basis = torch.exp(-theta * torch.abs(x.unsqueeze(0) - x.unsqueeze(1)))
        L = torch.linalg.cholesky(self.basis)
        self.c_bd = torch.cholesky_inverse(L)
        self.c_bk = self.c_bd @ self.w_dk

    def get_w_d(self, log_lambdas):
        return self.calc_w_d(self.w_dkq_sign, self.log_w_dkq, log_lambdas)

    def get_c_b(self, log_lambdas):
        w_d = self.get_w_d(log_lambdas)
        return self.c_bd @ w_d

    def calc_lambdas(self, covs, ns=None):
        if ns is None:
            lambdas = torch.linalg.solve(self.w_dk, covs)
        else:
            WD = self.w_dk.T * ns.unsqueeze(0)
            A = WD @ self.w_dk
            b = WD @ covs
            lambdas = torch.linalg.solve(A, b)
        return lambdas

class WkAligner(torch.nn.Module):
    def __init__(self, n_alleles, seq_length, max_k=None):
        super().__init__()
        self.n_alleles = n_alleles
        self.seq_length = seq_length
        self.ws = KrawtchoukPolynomials(n_alleles, seq_length, max_k)
        
    def set_data(self, cov, ns=None):
        self.cov = cov
        self.ns = ns
        if self.ns is None:
            self.ns = torch.ones_like(cov)
        

        print(self.cov)
        b = self.cov[1] / (self.cov[0] - self.cov[1])
        beta0 = np.log(self.cov[0]) + self.seq_length * (np.log(1 + self.n_alleles * b) - np.log(1 + b))
        beta1 = np.log(1 + self.n_alleles * b)
        k = torch.arange(self.ws.max_k + 1).to(dtype=torch.float)
        log_lambdas0 = beta0 - beta1 * k
        # log_lambdas0 =  - beta * k
        # log_lambdas0 = torch.zeros_like(k)
        # log_lambdas0[0] = 0.
        # print(log_lambdas0)
        # input()
        # log_lambdas0 = -torch.arange(self.ws.max_k + 1).to(dtype=torch.float)
        # print(log_lambdas0)
        self.log_lambdas = torch.nn.Parameter(log_lambdas0)

    def predict(self, log_lambdas):
        return(self.ws.get_w_d(log_lambdas))

    def calc_loss(self, log_lambdas):
        w_d = self.predict(log_lambdas)
        # print(w_d)
        print(w_d[:5].detach().numpy(), self.cov[:5].numpy())
        rmse = torch.sum(torch.square(self.cov - w_d) * self.ns) / self.ns.sum()
        return rmse

    def fit(self, n_iter=20, lr=0.5):
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)

        for i in range(n_iter):
            optimizer.zero_grad()
            # print(self.log_lambdas)
            loss = self.calc_loss(self.log_lambdas)
            loss.backward()
            optimizer.step()
            print(loss.detach().item())
        print(self.log_lambdas)
        print(self.predict(self.log_lambdas))
        print(self.cov)


class SquaredMatMulOperator(LinearOperator):
    def __init__(self, x1, x2):
        # Match batch dimensions
        self.x1 = x1
        self.x2 = x2
        super().__init__(x1, x2)

    def _size(self):
        return (self.x1.shape[0], self.x2.shape[1])

    def _matmul(self, rhs):
        u = torch.einsum("iy,yx,xm,zx,iz->im", self.x1, self.x2, rhs, self.x2, self.x1)
        return u

    def _t_matmul(self, rhs):
        return self._matmul(rhs)


class HammingDistanceCalculator(object):
    def __init__(self, seq_length, shift, scale):
        self.seq_length = seq_length
        self.shift = shift
        self.scale = scale

    def __call__(self, x1, x2):
        v = self.shift + self.scale * self.seq_length
        ones1 = torch.ones((x1.shape[0], 1), device=x1.device)
        ones2 = torch.ones((1, x2.shape[0]), device=x1.device)
        c = MatmulLinearOperator(v * ones1, ones2)
        return c + MatmulLinearOperator(x1, -self.scale * x2.T)


def inner_product(x1, x2, metric=None, diag=False):
    if diag:
        min_size = min(x1.shape[0], x2.shape[0])
        if metric is None:
            return (x1[:min_size, :] * x2[:min_size, :]).sum(1)
        else:
            return ((x1[:min_size, :] @ metric) * x2[:min_size, :]).sum(1)
    else:
        if metric is None:
            return x1 @ x2.T
        else:
            return x1 @ metric @ x2.T


def cov2corr(cov):
    v = 1 / torch.sqrt(torch.diag(cov))
    corr = v.unsqueeze(0) * cov * v.unsqueeze(1)
    return(corr)




import pandas as pd
import torch

from copy import deepcopy
from time import time
from tqdm import tqdm

from torch.optim import Adam, SGD
from gpytorch.models import ApproximateGP, ExactGP
from gpytorch.means import ZeroMean, ConstantMean
from gpytorch.mlls import MarginalLogLikelihood, VariationalELBO
from gpytorch.likelihoods import FixedNoiseGaussianLikelihood, _GaussianLikelihoodBase
from gpytorch.distributions import MultivariateNormal
from gpytorch.variational import (
    CholeskyVariationalDistribution,
    UnwhitenedVariationalStrategy,
)
from gpytorch.settings import (
    num_likelihood_samples,
    max_preconditioner_size,
    max_lanczos_quadrature_iterations,
    fast_pred_var,
    skip_posterior_variances,
    cg_tolerance,
    num_trace_samples,
)

from epik.utils import (
    get_tensor,
    to_numpy,
    split_training_test,
    encode_seqs,
    get_mut_effs_contrast_matrix,
    get_epistatic_coeffs_contrast_matrix,
)


class ExactMLL(MarginalLogLikelihood):
    def __init__(self, likelihood, model):
        if not isinstance(likelihood, _GaussianLikelihoodBase):
            raise RuntimeError("Likelihood must be Gaussian for exact inference")
        super(ExactMLL, self).__init__(likelihood, model)

    def _add_other_terms(self, res, params):
        # Add additional terms (SGPR / learned inducing points, heteroskedastic likelihood models)
        for added_loss_term in self.model.added_loss_terms():
            res = res.add(added_loss_term.loss(*params))

        # Add log probs of priors on the (functions of) parameters
        res_ndim = res.ndim
        for name, module, prior, closure, _ in self.model.named_priors():
            prior_term = prior.log_prob(closure(module))
            res.add_(prior_term.view(*prior_term.shape[:res_ndim], -1).sum(dim=-1))

        return res

    def forward(self, function_dist, target, *params):
        if not isinstance(function_dist, MultivariateNormal):
            raise RuntimeError(
                "ExactMarginalLogLikelihood can only operate on Gaussian random variables"
            )

        # Get the log prob of the marginal distribution
        output = self.likelihood(function_dist, *params)
        res = output.log_prob(target)
        res = self._add_other_terms(res, params)
        return res


class GPModel(ExactGP):
    def __init__(self, train_x, train_y, kernel, likelihood, train_mean=False):
        super(GPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean() if train_mean else ZeroMean()
        self.covar_module = kernel

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class GeneralizedGPModel(ApproximateGP):
    def __init__(self, train_x, kernel, train_mean=False):
        distribution = CholeskyVariationalDistribution(train_x.size(0))
        strategy = UnwhitenedVariationalStrategy(
            self, train_x, distribution, learn_inducing_locations=False
        )
        super(GeneralizedGPModel, self).__init__(strategy)
        self.mean_module = ConstantMean() if train_mean else ZeroMean()
        self.covar_module = kernel

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class _Epik(object):
    def __init__(
        self,
        kernel,
        device="cpu",
        train_mean=False,
        train_noise=False,
        preconditioner_size=0,
        cg_tol=1.0,
        num_trace_samples=50,
        max_n_lanczos_iterations=50,
        track_progress=False,
    ):
        self.kernel = kernel
        self.device = device
        self.train_mean = train_mean
        self.train_noise = train_noise

        self.preconditioner_size = preconditioner_size
        self.cg_tol = cg_tol
        self.num_trace_samples = num_trace_samples
        self.max_n_lanczos_iterations = max_n_lanczos_iterations
        self.track_progress = track_progress
        self.fit_time = 0
        self.training_history = []
        self.params_history = []

    def report_progress(self, pbar):
        if self.track_progress:
            allocated = torch.cuda.memory_allocated(device="cuda") / 1e6
            reserved = torch.cuda.memory_reserved(device="cuda") / 1e6
            report_dict = {
                "MLL": f"{self.mll:.3f}",
                "Mem(alloc/res)": f"{allocated:.2f}/{reserved:.2f}MB",
            }
            if hasattr(self, "scheduler"):
                report_dict["LR"] = f"{self.optimizer.param_groups[0]['lr']:.4f}"

            pbar.set_postfix(report_dict)

    def get_tensor(self, ndarray):
        return get_tensor(ndarray, device=self.device)

    def set_training_mode(self):
        self.gp.train()
        self.likelihood.train()

    def set_evaluation_mode(self):
        self.gp.eval()
        self.likelihood.eval()

    def set_data(self, X, y, y_var=None):
        """
        Set the training data for the model.

        Parameters
        ----------
        X : torch.Tensor
            A tensor of shape (n_sequences, n_features) containing the 
            one-hot encoded input sequences.

        y : torch.Tensor
            A tensor of shape (n_sequences,) containing the phenotypic 
            measurements corresponding to each sequence in `X`.

        y_var : torch.Tensor, optional
            A tensor of shape (n_sequences,) representing the variance 
            of the measurements in `y`. If `None`, it is assumed that 
            there is no uncertainty in the measurements.
        """

        self.X = self.get_tensor(X)
        self.y = self.get_tensor(y)
        self.y_var = (
            torch.zeros_like(self.y) if y_var is None else self.get_tensor(y_var)
        )

        self.define_likelihood()
        self.define_gp()

    def calc_mll(self):
        with max_preconditioner_size(self.preconditioner_size), cg_tolerance(
            self.cg_tol
        ), num_trace_samples(self.num_trace_samples), max_lanczos_quadrature_iterations(
            self.max_n_lanczos_iterations
        ):
            return self.mll_layer(self.gp(self.X), self.y)

    def training_step(self):
        self.optimizer.zero_grad()
        torch.cuda.empty_cache()
        neg_mll = -self.calc_mll()
        neg_mll.backward()
        self.optimizer.step()

        self.params = self.gp.state_dict()
        self.mll = -neg_mll.detach().item()
        self.params_history.append(self.params)
        self.training_history.append(self.mll)

        if not hasattr(self, "max_mll") or self.mll > self.max_mll:
            self.max_mll = self.mll
            self.max_params = deepcopy(self.params)

    def fit(self, n_iter=100, learning_rate=0.1, optimizer="Adam"):
        """
        Optimize model hyperparameters by maximizing the marginal likelihood.

        This process adjusts kernel parameters, as well as optional mean and 
        additional noise parameters, to improve the model's performance.

        Parameters
        ----------
        n_iter : int, optional (default=100)
            Number of iterations for the optimization process.

        learning_rate : float, optional (default=0.1)
            Learning rate for the optimizer.

        optimizer : str, optional (default="Adam")
            Optimizer to use for hyperparameter optimization. Options are 
            "Adam" or "SGD".

        Raises
        ------
        ValueError
            If the specified optimizer is not recognized.
        """
        if optimizer == "Adam":
            self.optimizer = Adam(self.gp.parameters(), lr=learning_rate)
        elif optimizer == "SGD":
            self.optimizer = SGD(self.gp.parameters(), lr=learning_rate)
        else:
            raise ValueError("Optimizer {} not recognized".format(optimizer))
        
        self.set_training_mode()

        t0 = time()
        pbar = range(n_iter)
        if n_iter > 1 and self.track_progress:
            pbar = tqdm(pbar, desc="Optimizing hyperparameters")

        for _ in pbar:
            try:
                self.training_step()
            except RuntimeError:
                self.kernel.use_keops = True
                self.training_step()

            if n_iter > 1:
                self.report_progress(pbar)

        self.fit_time = time() - t0

    @property
    def history(self):
        return pd.DataFrame({"mll": self.training_history})

    def get_params(self):
        return self.gp.state_dict()

    def save(self, fpath):
        """
        Save the model parameters to a file for future use.

        Parameters
        ----------
        fpath : str
            The file path where the model parameters will be saved.
        """
        torch.save(self.gp.state_dict(), fpath)

    def set_params(self, params):
        self.gp.load_state_dict(params)

    def load(self, fpath, **kwargs):
        """
        Load model parameters from a file.

        Parameters
        ----------
        fpath : str
            Path to the file containing the stored model parameters.

        **kwargs : dict, optional
            Additional arguments to pass to `torch.load` for loading the parameters.
        """
        params = torch.load(fpath, **kwargs)
        self.set_params(params)


class EpiK(_Epik):
    """
    Gaussian process regression model for inferring
    sequence-function relationships from experimental measurements
    using GPyTorch and KeOps backend.

    Parameters
    ----------
    kernel : epik.kernel.Kernel
        An instance of a kernel class that defines the covariance
        structure between pairs of sequences for Gaussian process regression.

    device : str, optional
        The device on which computations will be performed. Options are
        "cpu" or "cuda". Default is "cpu".

    train_mean : bool, optional
        Whether to optimize the mean function of the Gaussian Process.
        By default, it assumes a zero-mean function. Default is False.

    train_noise : bool, optional
        Whether to learn an additional noise parameter for the Gaussian Process.
        By default, it assumes that the provided error estimates are reliable.
        Default is False.

    preconditioner_size : int, optional
        The size of the preconditioner used to accelerate
        conjugate gradient convergence. By default, no
        preconditioner is computed. Default is 0.

    cg_tol : float, optional
        The tolerance level for the conjugate gradient solver. This parameter
        controls the precision of the solver. Lower values result in higher
        precision but may increase computation time. Default is 1.0.

    num_trace_samples : int, optional
        The number of samples used to estimate the trace of a matrix during
        computations. Increasing this value improves the accuracy of the trace
        estimation but increases computational cost. Default is 50.

    max_n_lanczos_iterations : int, optional
        The maximum number of Lanczos iterations to perform during matrix
        decompositions. Higher values may improve accuracy but increase
        computation time. Default is 50.

    track_progress : bool, optional
        Whether to display a progress bar during model fitting. Default is False.

    """
    def define_likelihood(self):
        self.likelihood = FixedNoiseGaussianLikelihood(
            noise=self.y_var, learn_additional_noise=self.train_noise
        )

        if self.device == "cuda":
            self.likelihood = self.likelihood.cuda()

    def get_gp(self, likelihood, x=None, y=None):
        gp = GPModel(x, y, self.kernel, likelihood, train_mean=self.train_mean)

        if self.device == "cuda":
            gp = gp.cuda()

        return gp

    def define_gp(self):
        self.gp = self.get_gp(self.likelihood, self.X, self.y)
        self.mll_layer = ExactMLL(self.likelihood, self.gp)

    def get_posterior(self, X, calc_variance=False, calc_covariance=False):
        self.set_evaluation_mode()
        X = self.get_tensor(X)

        with torch.no_grad(), max_preconditioner_size(self.preconditioner_size):
            if calc_covariance:
                f = self.gp(X)
            elif calc_variance:
                with fast_pred_var():
                    f = self.gp(X)
            else:
                with skip_posterior_variances():
                    f = self.gp(X)
        return f

    def pred_to_df(self, res, calc_variance=False, labels=None):
        if calc_variance:
            m, x = res
            if len(x.shape) == 1:
                var = x
            else:
                var = x.diag()
            sd = to_numpy(torch.sqrt(var))
            m = to_numpy(m)
            result = pd.DataFrame(
                {
                    "coef": m,
                    "stderr": sd,
                    "lower_ci": m - 2 * sd,
                    "upper_ci": m + 2 * sd,
                },
                index=labels,
            )
        else:
            result = pd.DataFrame({"coef": to_numpy(res)}, index=labels)
        return result

    def predict(self, X, calc_variance=False, labels=None):
        """
        Function to make phenotypic predictions under the
        Gaussian process model

        Parameters
        ----------
        X : torch.Tensor of shape (n_sequences, n_features)
            Tensor containing the one-hot encoding of the
            sequences to make predictions

        calc_variance : bool (False)
            Option to compute the posterior variance in addition
            to the posterior mean reported by default

        labels : array-like of shape (n_sequences,) or None
            Sequence labels to use as rownames in the output
            pd.DataFrame

        Returns
        -------
        output : pd.DataFrame of shape (n_sequences, 1 or 4)
            DataFrame containing phenotypic predictions at the
            desired sequences. If `calc_variance=True`, posterior
            standard deviations and 95% credible interval
            bounds are added

        """
        t0 = time()
        self.set_evaluation_mode()
        X = self.get_tensor(X)
        f = self.get_posterior(X, calc_variance=calc_variance)

        res = (f.mean, f.variance) if calc_variance else f.mean
        df = self.pred_to_df(res, calc_variance=calc_variance, labels=labels)
        self.pred_time = time() - t0
        return df

    def make_contrasts(self, contrast_matrix, X, calc_variance=False):
        """
        Compute phenotypic contrasts across sets of genotypes
        using the Gaussian process model.

        Parameters
        ----------
        contrast_matrix : torch.Tensor of shape (n_contrasts, n_sequences)
            A tensor representing the linear combinations of sequences 
            encoded by `X` to compute the posterior distribution of 
            the contrasts.

        X : torch.Tensor of shape (n_sequences, n_features)
            A tensor containing the one-hot encoded sequences for 
            which predictions are to be made.

        calc_variance : bool, optional (default=False)
            If True, computes the posterior (co)-variance in addition 
            to the posterior mean.

        Returns
        -------
        output : torch.Tensor or tuple of torch.Tensor
            If `calc_variance=False`, returns a tensor containing the 
            phenotypic predictions for the desired sequences. 
            If `calc_variance=True`, returns a tuple where the first 
            element is the phenotypic predictions and the second element 
            is the covariance matrix of the posterior contrasts.
        """
        t0 = time()
        self.set_evaluation_mode()

        B = self.get_tensor(contrast_matrix)
        f = self.get_posterior(X, calc_covariance=calc_variance)

        res = B @ f.mean
        if calc_variance:
            res = res, (B @ f.covariance_matrix @ B.T)

        self.contrast_time = time() - t0
        return res

    def predict_contrasts(
        self, contrast_matrix, alleles, calc_variance=False, max_size=100
    ):
        n = contrast_matrix.shape[0]
        n_chunks = int(n / max_size) + 1
        results = []
        chunks = range(n_chunks)
        if self.track_progress:
            chunks = tqdm(chunks, total=n_chunks)
        for i in chunks:
            df = contrast_matrix.iloc[i * max_size : (i + 1) * max_size, :]
            seqs, labels = df.columns, df.index
            X = encode_seqs(seqs, alphabet=alleles)
            C = torch.Tensor(df.values)
            res = self.make_contrasts(C, X, calc_variance=calc_variance)
            result = self.pred_to_df(res, calc_variance=calc_variance, labels=labels)
            results.append(result)
        results = pd.concat(results)
        return results

    def predict_mut_effects(self, seq0, alleles, calc_variance=False, max_size=100):
        """
        Predict the effects of single mutations on the phenotype
        using the Gaussian process model.

        Parameters
        ----------
        seq0 : str
            The reference sequence for which single mutation effects
            are to be predicted.

        alleles : list of str
            A list of possible alleles for each position in the sequence.

        calc_variance : bool, optional (default=False)
            If True, computes the posterior variance in addition to the
            posterior mean.

        max_size : int, optional (default=100)
            The maximum number of contrasts to process in a single batch.
            Larger values may increase memory usage.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the predicted effects of single mutations
            on the phenotype. If `calc_variance=True`, the DataFrame includes
            posterior standard deviations and 95% credible interval bounds.
        """
        contrast_matrix = get_mut_effs_contrast_matrix(seq0, alleles)
        return self.predict_contrasts(
            contrast_matrix, alleles, calc_variance=calc_variance, max_size=max_size
        )

    def predict_epistatic_coeffs(
        self, seq0, alleles, calc_variance=False, max_size=100
    ):
        """
        Compute epistatic coefficients across sets of genotypes
        using the Gaussian process model.

        Parameters
        ----------
        seq0 : str
            The reference sequence for which epistatic coefficients
            are to be predicted.

        alleles : list of str
            A list of possible alleles for each position in the sequence.

        calc_variance : bool, optional (default=False)
            If True, computes the posterior (co)-variance in addition
            to the posterior mean.

        max_size : int, optional (default=100)
            The maximum number of contrasts to process in a single batch.
            Larger values may increase memory usage.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the predicted epistatic coefficients
            for the desired sequences. If `calc_variance=True`, the DataFrame
            includes posterior standard deviations and 95% credible interval
            bounds.
        """
        contrast_matrix = get_epistatic_coeffs_contrast_matrix(seq0, alleles)
        return self.predict_contrasts(
            contrast_matrix, alleles, calc_variance=calc_variance, max_size=max_size
        )

    def get_prior(self, X, sigma2):
        likelihood = FixedNoiseGaussianLikelihood(noise=sigma2 * torch.ones(X.shape[0]))
        gp = GPModel(None, None, self.kernel, likelihood, train_mean=self.train_mean)
        prior = gp.forward(X)
        return prior

    def simulate(self, X, n=1, sigma2=1e-4):
        """
        Sample random sequence-function relationships from the prior
        evaluated at the input sequences

        Parameters
        ----------
        X : torch.Tensor of shape (n_sequence, n_features)
            Tensor containing the one-hot encoding of the
            sequences to make predictions
        n : int (1)
            Number of sequence-function relationships to
            sample from the prior
        sigma2 : float (1e-4)
            Additional random noise to add to the
            simulated landscapes

        Returns
        -------
        y : torch.Tensor of shape (n_sequences, n)
            Tensor containing the simulated landscapes
            evaluated in the input sequences
        """

        with torch.no_grad(), max_preconditioner_size(self.preconditioner_size):
            prior = self.get_prior(X, sigma2=sigma2)
            v = torch.zeros(n)
            y = prior.rsample(v.size())

        return y

    def simulate_dataset(self, X, sigma=0, ptrain=0.8):
        """
        Simulate a dataset by sampling random sequence-function relationships 
        from the prior and splitting the data into training and test sets.

        Parameters
        ----------
        X : torch.Tensor of shape (n_sequence, n_features)
            Tensor containing the one-hot encoding of the sequences 
            to make predictions.

        sigma : float, optional (default=0)
            Standard deviation of the noise to add to the training data. 
            If `sigma=0`, no noise is added.

        ptrain : float, optional (default=0.8)
            Proportion of the data to include in the training set. 
            The remaining data will be used as the test set.

        Returns
        -------
        tuple
            A tuple containing:
            - train_x : torch.Tensor
            Training set input features.
            - train_y : torch.Tensor
            Training set target values.
            - test_x : torch.Tensor
            Test set input features.
            - test_y : torch.Tensor
            Test set target values.
            - train_y_var : torch.Tensor
            Variance of the training set target values.
        """
        y_true = self.simulate(X, n=1).flatten()
        y_true = y_true / y_true.std()

        splits = split_training_test(X, y_true, y_var=None, ptrain=ptrain)
        train_x, train_y, test_x, test_y, train_y_var = splits
        if sigma > 0:
            train_y = torch.normal(train_y, sigma)
            train_y_var = torch.full_like(train_y, sigma**2)

        return (train_x, train_y, test_x, test_y, train_y_var)


class GeneralizedEpiK(_Epik):
    def __init__(self, kernel, likelihood, **kwargs):
        super(self).__init__(kernel, **kwargs)
        self.likelihood_function = likelihood

    def get_likelihood(self, y_var, train_noise):
        likelihood = self.likelihood_function(y_var, train_noise)

        if self.device is not None:
            likelihood = likelihood.cuda()
        return likelihood

    def define_negative_loss(self):
        self.calc_negative_loss = VariationalELBO(
            self.likelihood, self.gp, self.y.numel()
        )

    def define_model(self):
        self.gp = GeneralizedGPModel(
            self.X,
            self.kernel,
            train_mean=self.train_mean,
            device=self.device,
            n_devices=self.n_devices,
        )
        if self.device is not None:
            self.gp = self.gp.cuda()

    def predict(self, X, nsamples=100):
        y_var = self.likelihood.second_noise * torch.ones(X.shape[0])
        likelihood = self.get_likelihood(y_var, train_noise=False)

        with torch.no_grad(), num_likelihood_samples(nsamples):
            phi = self.gp(X)
            y = likelihood(phi)
            yhat, y_var = y.mean.mean(0), y.variance.mean(0)
            phi = phi.mean.detach()
            return (phi, yhat, y_var)
